"""
Detect variants in reads.
"""
import logging
from collections import defaultdict, Counter
from typing import Iterable, Iterator, List, Optional

from .core import Read, ReadSet, NumericSampleIds
from .bam import SampleBamReader, MultiBamReader, BamReader
from .align import edit_distance, edit_distance_affine_gap
from ._variants import _iterate_cigar


logger = logging.getLogger(__name__)


class ReadSetError(Exception):
    pass


class ReadSetReader:
    """
    Associate VCF variants with BAM reads.

    A VCF file contains variants, and a BAM file contain reads, but the
    information which read contains which variant is not available. This
    class re-discovers the variants in each read, using the
    knowledge in the VCF of where they should occur.
    """

    def __init__(
        self,
        paths: List[str],
        reference: Optional[str],
        numeric_sample_ids: NumericSampleIds,
        mapq_threshold: int = 20,
        overhang: int = 10,
        affine: int = False,
        gap_start: int = 10,
        gap_extend: int = 7,
        default_mismatch: int = 15,
    ):
        """
        paths -- list of BAM paths
        reference -- path to reference FASTA (can be None)
        numeric_sample_ids -- ??
        mapq_threshold -- minimum mapping quality
        overhang -- extend alignment by this many bases to left and right
        affine -- use affine gap costs
        gap_start, gap_extend, default_mismatch -- parameters for affine gap cost alignment
        """
        self._mapq_threshold = mapq_threshold
        self._numeric_sample_ids = numeric_sample_ids
        self._use_affine = affine
        self._gap_start = gap_start
        self._gap_extend = gap_extend
        self._default_mismatch = default_mismatch
        self._overhang = overhang
        self._paths = paths
        self._reader: BamReader
        if len(paths) == 1:
            self._reader = SampleBamReader(paths[0], reference=reference)
        else:
            self._reader = MultiBamReader(paths, reference=reference)

    @property
    def n_paths(self):
        return len(self._paths)

    def read(self, chromosome, variants, sample, reference, regions=None) -> ReadSet:
        """
        Detect alleles and return a ReadSet object containing reads representing
        the given variants.

        If a reference is provided (reference is not None), alleles are
        detected by re-aligning sections of the query to the REF and ALT
        sequence extended a few bases to the left and right.

        If reference is None, alleles are detected by inspecting the
        existing alignment (via the CIGAR).

        chromosome -- name of chromosome to work on
        variants -- list of vcf.VcfVariant objects
        sample -- name of sample to work on. If None, read group information is
            ignored and all reads in the file are used.
        reference -- reference sequence of the given chromosome (or None)
        regions -- list of start,end tuples (end can be None)
        """
        # Since variants are identified by position, positions must be unique.
        if __debug__ and variants:
            varposc = Counter(variant.position for variant in variants)
            pos, count = varposc.most_common()[0]
            assert count == 1, f"Position {pos} occurs more than once in variant list."

        alignments = self._usable_alignments(chromosome, sample, regions)
        reads = self._alignments_to_reads(alignments, variants, sample, reference)
        grouped_reads = self._group_paired_reads(reads)
        readset = self._make_readset_from_grouped_reads(grouped_reads)
        return readset

    @staticmethod
    def _make_readset_from_grouped_reads(groups: Iterable[List[Read]]) -> ReadSet:
        read_set = ReadSet()
        for group in groups:
            read_set.add(merge_reads(*group))
        return read_set

    @staticmethod
    def _group_paired_reads(reads: Iterable[Read]) -> Iterator[List[Read]]:
        """
        Group reads into paired-end read pairs. Uses name, source_id and sample_id
        as grouping key.

        TODO
        Grouping by name should be sufficient since the SAM spec states:
        "Reads/segments having identical QNAME are regarded to come from the same template."
        """
        groups = defaultdict(list)
        for read in reads:
            groups[(read.source_id, read.name, read.sample_id)].append(read)
        for group in groups.values():
            if len(group) > 2:
                raise ReadSetError(
                    f"Read name {group[0].name!r} occurs more than twice in the input file"
                )
            yield group

    def _usable_alignments(self, chromosome, sample, regions=None):
        """
        Retrieve usable (suficient mapping quality, not secondary etc.)
        alignments from the alignment file
        """
        if regions is None:
            regions = [(0, None)]
        for s, e in regions:
            for alignment in self._reader.fetch(
                reference=chromosome, sample=sample, start=s, end=e
            ):
                # TODO handle additional alignments correctly!
                # find out why they are sometimes overlapping/redundant
                if (
                    alignment.bam_alignment.flag & 2048 != 0
                    or alignment.bam_alignment.mapping_quality < self._mapq_threshold
                    or alignment.bam_alignment.is_secondary
                    or alignment.bam_alignment.is_unmapped
                    or alignment.bam_alignment.is_duplicate
                ):
                    continue
                yield alignment

    def has_reference(self, chromosome):
        return self._reader.has_reference(chromosome)

    def _alignments_to_reads(self, alignments, variants, sample, reference):
        """
        Convert BAM alignments to Read objects.

        If reference is not None, alleles are detected through re-alignment.

        Yield Read objects.
        """
        # FIXME hard-coded zero
        numeric_sample_id = 0 if sample is None else self._numeric_sample_ids[sample]
        if reference is not None:
            # Copy the pyfaidx.FastaRecord into a str for faster access
            reference = reference[:]
            normalized_variants = variants
        else:
            normalized_variants = [variant.normalized() for variant in variants]

        i = 0  # index into variants
        for alignment in alignments:
            # Skip variants that are to the left of this read
            while (
                i < len(normalized_variants)
                and normalized_variants[i].position < alignment.bam_alignment.reference_start
            ):
                i += 1

            try:
                barcode = alignment.bam_alignment.get_tag("BX")
            except KeyError:
                barcode = ""

            read = Read(
                alignment.bam_alignment.qname,
                alignment.bam_alignment.mapq,
                alignment.source_id,
                numeric_sample_id,
                alignment.bam_alignment.reference_start,
                barcode,
            )

            if reference is None:
                detected = self.detect_alleles(normalized_variants, i, alignment.bam_alignment)
            else:
                detected = self.detect_alleles_by_alignment(
                    variants,
                    i,
                    alignment.bam_alignment,
                    reference,
                    self._overhang,
                    self._use_affine,
                    self._gap_start,
                    self._gap_extend,
                    self._default_mismatch,
                )
            for j, allele, quality in detected:
                read.add_variant(variants[j].position, allele, quality)
            if read:  # At least one variant covered and detected
                yield read

    @staticmethod
    def detect_alleles(variants, j, bam_read):
        """
        Detect the correct alleles of the variants that are covered by the
        given bam_read.

        Yield tuples (index, allele, quality), where index is into the variants list.

        variants -- list of variants (VcfVariant objects)
        j -- index of the first variant (in the variants list) to check
        """
        ref_pos = bam_read.reference_start  # position relative to reference
        query_pos = 0  # position relative to read

        seen_positions = set()
        for cigar_op, length in bam_read.cigartuples:
            # Skip variants that come before this region
            while j < len(variants) and variants[j].position < ref_pos:
                j += 1

            # The mapping of CIGAR operators to numbers is:
            # MIDNSHPX= => 012345678
            if cigar_op in (0, 7, 8):  # M, X, = operators (match)
                # Iterate over all variants that are in this region
                while j < len(variants) and variants[j].position < ref_pos + length:
                    if (
                        len(variants[j].reference_allele)
                        == len(variants[j].alternative_allele)
                        == 1
                    ):
                        # Variant is a SNV
                        offset = variants[j].position - ref_pos
                        base = bam_read.query_sequence[query_pos + offset]
                        if base == variants[j].reference_allele:
                            allele = 0
                        elif base == variants[j].alternative_allele:
                            allele = 1
                        else:
                            allele = None
                        if allele is not None:
                            # TODO
                            # Fix this: we can actually have indel and SNV
                            # calls at identical positions. For now, ignore the
                            # second variant.
                            if variants[j].position in seen_positions:
                                logger.debug(
                                    "Found two variants at identical positions."
                                    " Ignoring the second one: %s",
                                    variants[j],
                                )
                            else:
                                # Do not use bam_read.qual here as it is extremely slow.
                                # If we ever decide to be compatible with older pysam
                                # versions, cache bam_read.qual somewhere - do not
                                # access it within this loop (3x slower otherwise).
                                if bam_read.query_qualities:
                                    qual = bam_read.query_qualities[query_pos + offset]
                                else:
                                    qual = 30  # TODO
                                yield (j, allele, qual)
                                seen_positions.add(variants[j].position)
                    elif len(variants[j].reference_allele) == 0:
                        assert len(variants[j].alternative_allele) > 0
                        # This variant is an insertion. Since we are in a region of
                        # matches, the insertion was *not* observed (reference allele).
                        qual = 30  # TODO average qualities of "not inserted" bases?
                        yield (j, 0, qual)
                        seen_positions.add(variants[j].position)
                    elif len(variants[j].alternative_allele) == 0:
                        assert len(variants[j].reference_allele) > 0
                        # This variant is a deletion that was not observed.
                        # Add it only if the next variant is not located 'within'
                        # the deletion.
                        deletion_end = variants[j].position + len(variants[j].reference_allele)
                        if not (j + 1 < len(variants) and variants[j + 1].position < deletion_end):
                            qual = 30  # TODO
                            yield (j, 0, qual)
                            seen_positions.add(variants[j].position)
                        else:
                            logger.info(
                                "Skipped a deletion overlapping another variant at pos. %d",
                                variants[j].position,
                            )
                            # Also skip all variants that this deletion overlaps
                            while j + 1 < len(variants) and variants[j + 1].position < deletion_end:
                                j += 1
                            # One additional j += 1 is done below
                    else:
                        assert False, f"Strange variant: {variants[j]}"
                    j += 1
                query_pos += length
                ref_pos += length
            elif cigar_op == 1:  # I operator (insertion)
                if (
                    j < len(variants)
                    and variants[j].position == ref_pos
                    and len(variants[j].reference_allele) == 0
                    and variants[j].alternative_allele
                    == bam_read.query_sequence[query_pos : query_pos + length]
                ):
                    qual = 30  # TODO
                    assert variants[j].position not in seen_positions
                    yield (j, 1, qual)
                    seen_positions.add(variants[j].position)
                    j += 1
                query_pos += length
            elif cigar_op == 2:  # D operator (deletion)
                # We only check the length of the deletion, not the sequence
                # that gets deleted since we don’t have the reference available.
                # (We could parse the MD tag if it exists.)
                if (
                    j < len(variants)
                    and variants[j].position == ref_pos
                    and len(variants[j].alternative_allele) == 0
                    and len(variants[j].reference_allele) == length
                ):
                    qual = 30  # TODO
                    deletion_end = variants[j].position + len(variants[j].reference_allele)
                    if not (j + 1 < len(variants) and variants[j + 1].position < deletion_end):
                        qual = 30  # TODO
                        assert variants[j].position not in seen_positions
                        yield (j, 1, qual)
                        seen_positions.add(variants[j].position)
                    else:
                        logger.info(
                            "Skipped a deletion overlapping another variant at pos. %d",
                            variants[j].position,
                        )
                        # Also skip all variants that this deletion overlaps
                        while j + 1 < len(variants) and variants[j + 1].position < deletion_end:
                            j += 1
                        # One additional j += 1 is done below
                    j += 1
                ref_pos += length
            elif cigar_op == 3:  # N operator (reference skip)
                ref_pos += length
            elif cigar_op == 4:  # S operator (soft clipping)
                query_pos += length
            elif cigar_op == 5 or cigar_op == 6:  # H or P (hard clipping or padding)
                pass
            else:
                logger.error("Unsupported CIGAR operation: %d", cigar_op)
                raise ValueError(f"Unsupported CIGAR operation: {cigar_op}")

    @staticmethod
    def split_cigar(cigar, i, consumed):
        """
        Split a CIGAR into two parts. i and consumed describe the split position.
        i is the element of the cigar list that should be split, and consumed says
        at how many operations to split within that element.

        The CIGAR is given as a list of (operation, length) pairs.

        i -- split at this index in cigar list
        consumed -- how many cigar ops at cigar[i] are to the *left* of the
            split position

        Return a tuple (left, right).

        Example:
        Assume the cigar is 3M 1D 6M 2I 4M.
        With i == 2 and consumed == 5, the cigar is split into
        3M 1D 5M and 1M 2I 4M.
        """
        middle_op, middle_length = cigar[i]
        assert consumed <= middle_length
        if consumed > 0:
            left = cigar[:i] + [(middle_op, consumed)]
        else:
            left = cigar[:i]
        if consumed < middle_length:
            right = [(middle_op, middle_length - consumed)] + cigar[i + 1 :]
        else:
            right = cigar[i + 1 :]
        return left, right

    @staticmethod
    def cigar_prefix_length(cigar, reference_bases):
        """
        Given a prefix of length reference_bases relative to the reference, how
        long is the prefix of the read? In other words: If reference_bases on
        the reference are consumed, how many bases on the query does that
        correspond to?

        If the position is within or at the end of an insertion (which do not
        consume bases on the reference), then the number of bases up to the
        beginning of the insertion is reported.

        Return a pair (reference_bases, query_bases) where the value for
        reference_bases may be smaller than the requested one if the CIGAR does
        not cover enough reference bases.

        Reference skips (N operators) are treated as the end of the read. That
        is, no positions beyond a reference skip are reported.
        """
        ref_pos = 0
        query_pos = 0
        for op, length in cigar:
            if op in (0, 7, 8):  # M, X, =
                ref_pos += length
                query_pos += length
                if ref_pos >= reference_bases:
                    return (reference_bases, query_pos + reference_bases - ref_pos)
            elif op == 2:  # D
                ref_pos += length
                if ref_pos >= reference_bases:
                    return (reference_bases, query_pos)
            elif op == 1:  # I
                query_pos += length
            elif op == 4 or op == 5:  # soft or hard clipping
                pass
            elif op == 3:  # N
                # Always stop at reference skips
                return (reference_bases, query_pos)
            else:
                assert False, "unknown CIGAR operator"
        assert ref_pos < reference_bases
        return (ref_pos, query_pos)

    @staticmethod
    def realign(
        variant,
        bam_read,
        cigartuples,
        i,
        consumed,
        query_pos,
        reference,
        overhang,
        use_affine,
        gap_start,
        gap_extend,
        default_mismatch,
    ):
        """
        Realign a read to the two alleles of a single variant.
        i and consumed describe where to split the cigar into a part before the
        variant position and into a part starting at the variant position, see split_cigar().

        variant -- VcfVariant
        bam_read -- the AlignedSegment
        cigartuples -- the AlignedSegment.cigartuples property (accessing it is expensive, so re-use it)
        i, consumed -- see split_cigar method
        query_pos -- index of the query base that is at the variant position
        reference -- the reference as a str-like object (full chromosome)
        overhang -- extend alignment by this many bases to left and right
        use_affine -- if true, use affine gap costs for realignment
        gap_start, gap_extend -- if affine_gap=true, use these parameters for affine gap cost alignment
        default_mismatch -- if affine_gap=true, use this as mismatch cost in case no base qualities are in bam
        """
        # Do not process symbolic alleles like <DEL>, <DUP>, etc.
        if variant.alternative_allele.startswith("<"):
            return None, None

        left_cigar, right_cigar = ReadSetReader.split_cigar(cigartuples, i, consumed)

        left_ref_bases, left_query_bases = ReadSetReader.cigar_prefix_length(
            left_cigar[::-1], overhang
        )
        right_ref_bases, right_query_bases = ReadSetReader.cigar_prefix_length(
            right_cigar, len(variant.reference_allele) + overhang
        )

        assert variant.position - left_ref_bases >= 0
        assert variant.position + right_ref_bases <= len(reference)

        query = bam_read.query_sequence[
            query_pos - left_query_bases : query_pos + right_query_bases
        ]
        ref = reference[variant.position - left_ref_bases : variant.position + right_ref_bases]
        alt = (
            reference[variant.position - left_ref_bases : variant.position]
            + variant.alternative_allele
            + reference[
                variant.position
                + len(variant.reference_allele) : variant.position
                + right_ref_bases
            ]
        )

        if use_affine:
            assert gap_start is not None
            assert gap_extend is not None
            assert default_mismatch is not None

            # get base qualities if present (to be used as mismatch costs)
            base_qualities = [default_mismatch] * len(query)
            # if bam_read.query_qualities != None:
            #    base_qualities = bam_read.query_qualities[query_pos-left_query_bases:query_pos+right_query_bases]

            # compute edit dist. with affine gap costs using base qual. as mismatch cost
            distance_ref = edit_distance_affine_gap(
                query, ref, base_qualities, gap_start, gap_extend
            )
            distance_alt = edit_distance_affine_gap(
                query, alt, base_qualities, gap_start, gap_extend
            )
            base_qual_score = abs(distance_ref - distance_alt)
        else:
            base_qual_score = 30
            distance_ref = edit_distance(query, ref)
            distance_alt = edit_distance(query, alt)

        if distance_ref < distance_alt:
            return 0, base_qual_score  # detected REF
        elif distance_ref > distance_alt:
            return 1, base_qual_score  # detected ALT
        else:
            return None, None  # cannot decide

    @staticmethod
    def detect_alleles_by_alignment(
        variants,
        j,
        bam_read,
        reference,
        overhang=10,
        use_affine=False,
        gap_start=None,
        gap_extend=None,
        default_mismatch=None,
    ):
        """
        Detect which alleles the given bam_read covers. Detect the correct
        alleles of the variants that are covered by the given bam_read.

        Yield tuples (position, allele, quality).

        variants -- list of variants (VcfVariant objects)
        j -- index of the first variant (in the variants list) to check
        """
        # Accessing bam_read.cigartuples is expensive, do it only once
        cigartuples = bam_read.cigartuples

        # For the same reason, the following check is here instad of
        # in the _usable_alignments method
        if not cigartuples:
            return

        for index, i, consumed, query_pos in _iterate_cigar(variants, j, bam_read, cigartuples):
            allele, quality = ReadSetReader.realign(
                variants[index],
                bam_read,
                cigartuples,
                i,
                consumed,
                query_pos,
                reference,
                overhang,
                use_affine,
                gap_start,
                gap_extend,
                default_mismatch,
            )
            if allele in (0, 1):
                yield (index, allele, quality)  # TODO quality???

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        self._reader.close()


def merge_two_reads(read1: Read, read2: Read) -> Read:
    """
    Merge two reads *that belong to the same haplotype* (such as the two
    ends of a paired-end read) into a single Read. Overlaps are allowed.
    """
    assert read1.is_sorted()
    assert read2.is_sorted()
    if read2:
        result = Read(
            read1.name,
            read1.mapqs[0],
            read1.source_id,
            read1.sample_id,
            read1.reference_start,
            read1.BX_tag,
        )
        result.add_mapq(read2.mapqs[0])
    else:
        return read1

    i1 = 0
    i2 = 0

    def add1():
        result.add_variant(read1[i1].position, read1[i1].allele, read1[i1].quality)

    def add2():
        result.add_variant(read2[i2].position, read2[i2].allele, read2[i2].quality)

    while i1 < len(read1) or i2 < len(read2):
        if i1 == len(read1):
            add2()
            i2 += 1
            continue
        if i2 == len(read2):
            add1()
            i1 += 1
            continue
        variant1 = read1[i1]
        variant2 = read2[i2]
        if variant2.position < variant1.position:
            add2()
            i2 += 1
        elif variant2.position > variant1.position:
            add1()
            i1 += 1
        else:
            # Variant on self-overlapping read pair
            assert read1[i1].position == read2[i2].position
            # If both alleles agree, merge into single variant and add up qualities
            if read1[i1].allele == read2[i2].allele:
                quality = read1[i1].quality + read2[i2].quality
                result.add_variant(read1[i1].position, read1[i1].allele, quality)
            else:
                # Otherwise, take variant with highest base quality and discard the other.
                if read1[i1].quality >= read2[i2].quality:
                    add1()
                else:
                    add2()
            i1 += 1
            i2 += 1
    return result


def merge_reads(*reads: Read) -> Read:
    """
    Merge multiple reads that belong to the same haplotype into a single Read.

    If the iterable is empty, a ValueError is raised.

    This 'naive' version just calls merge_two_reads repeatedly on all the reads.

    # TODO
    # The actual challenge is dealing with conflicts in variants covered by
    # more than one read. A solution would be to not merge if there are any
    # (or too many) conflicts and let the main algorithm deal with it.
    """
    it = iter(reads)
    try:
        read = next(it)
    except StopIteration:
        raise ValueError("no reads to merge")
    assert read.is_sorted()
    for partner in it:
        read = merge_two_reads(read, partner)
    return read



import itertools
import math

from whatshap.polyphase_solver import ClusterEditingSolver, scoreReadset
from whatshap.testhelpers import string_to_readset


def test_similarities1():
    reads = """
    001001
    110101
    """
    readset = string_to_readset(reads)
    similarities = scoreReadset(readset, 4, 2, 0.06)

    assert not math.isnan(similarities.get(0, 1))
    assert similarities.get(0, 1) < -6.0


def test_similarities2():
    reads = """
    00000
    00000
    00000
    00000
    11111
    11111
    10101
    10101
    """
    readset = string_to_readset(reads)
    similarities = scoreReadset(readset, 4, 4, 0.06)

    assert similarities.get(0, 1) > 1.0
    assert similarities.get(0, 1) == similarities.get(0, 2) == similarities.get(0, 3)
    assert similarities.get(0, 1) == similarities.get(1, 2) == similarities.get(1, 3)
    assert similarities.get(0, 4) < -8.0
    assert similarities.get(0, 5) < -8.0
    assert similarities.get(0, 6) < -1.0
    assert similarities.get(0, 7) < -1.0
    assert similarities.get(4, 5) > 1.0
    assert similarities.get(4, 6) < -1.0


def test_similarities3():
    reads = """
    00000
    00000
    00000
    00000
    11111
    11111
    10101
    10101
    """
    readset = string_to_readset(reads)
    similarities = scoreReadset(readset, 4, 2, 0.06)

    assert similarities.get(0, 1) > 5.0
    assert similarities.get(0, 1) == similarities.get(0, 2) == similarities.get(0, 3)
    assert similarities.get(0, 1) == similarities.get(1, 2) == similarities.get(1, 3)
    assert similarities.get(0, 4) < -8.0
    assert similarities.get(0, 5) < -8.0
    assert similarities.get(0, 6) < -1.0
    assert similarities.get(0, 7) < -1.0
    assert similarities.get(4, 5) > 1.0
    assert similarities.get(4, 6) > 1.0


def test_similarities4():
    reads = """
    000
     000
      000
    111
     111
      101
     110
    """
    readset = string_to_readset(reads)
    similarities = scoreReadset(readset, 2, 2, 0.06)

    assert similarities.get(0, 1) > 2.0
    assert similarities.get(0, 2) == 0.0
    assert similarities.get(1, 2) > 2.0
    assert similarities.get(0, 3) < -5.0 < similarities.get(1, 3) < 0.0 == similarities.get(2, 3)
    assert similarities.get(4, 6) > similarities.get(4, 5) > 0.0


def test_similarities5():
    reads = """
    000
     000
      000
    111
     111
      101
     110
    """
    readset = string_to_readset(reads)
    similarities = scoreReadset(readset, 2, 3, 0.06)

    assert similarities.get(0, 1) > 1.0
    assert similarities.get(0, 2) == 0.0
    assert similarities.get(1, 2) > 0.5
    assert similarities.get(0, 3) < -5.0 < similarities.get(1, 3) < 0.0 == similarities.get(2, 3)
    assert 0.0 > similarities.get(4, 6) > similarities.get(4, 5)


def test_clusterediting1():

    reads = """
        110000010111
        1100000101
         1000 01
         00 0 0 010
         1000001 11
          1111101
          0 10010 1
           0000 010
           1110
           0000 011
            000  10
            0001011
            0  10110
            00010111
            000 0000
        """

    # construct a ReadSet
    readset = string_to_readset(reads)

    # compute similarities
    similarities = scoreReadset(readset, 3, 3, 0.06)

    # run cluster editing
    clusterediting = ClusterEditingSolver(similarities, False)
    readpartitioning = clusterediting.run()

    print("computed clusters: ", readpartitioning)

    # make sure each read occurs only once
    read_ids = list(itertools.chain.from_iterable(readpartitioning))
    duplicates = set([r for r in read_ids if read_ids.count(r) > 1])
    assert len(duplicates) == 0

    assert any(all(x in c for x in [0, 1, 2, 4, 9, 11, 13]) for c in readpartitioning)
    assert any(all(x in c for x in [3, 7, 10, 14]) for c in readpartitioning)
    assert any(all(x in c for x in [5, 8]) for c in readpartitioning)


def test_clusterediting2():

    reads = """
        000000 00 0 00000 0000 0
             1111 11111
               000 00000 0000000
               111111111
                 1000000000
                  0 00000
                    11111
                    1 1 1111 1111111111
                    111111111111
        """

    # construct a ReadSet
    readset = string_to_readset(reads)

    # compute similarities
    similarities = scoreReadset(readset, 3, 2, 0.06)

    # run cluster editing
    clusterediting = ClusterEditingSolver(similarities, False)
    readpartitioning = clusterediting.run()

    assert any(all(x in c for x in [0, 2, 4, 5]) for c in readpartitioning)
    assert any(all(x in c for x in [1, 3, 6, 7, 8]) for c in readpartitioning)


def test_clusterediting3():

    reads = """
        000000 00 0 00000 0000 0
             1111 11111
               000 00000 0000000
               111111111
                 1000000000
                  0 00000
                    11111
                    1 1 1111 1111111111
                    111111111111
        """

    # construct a ReadSet
    readset = string_to_readset(reads)

    # compute similarities
    similarities = scoreReadset(readset, 3, 2, 0.06)

    # run cluster editing
    clusterediting = ClusterEditingSolver(similarities, False)
    readpartitioning = clusterediting.run()

    assert any(all(x in c for x in [0, 2, 4, 5]) for c in readpartitioning)
    assert any(all(x in c for x in [1, 3, 6, 7, 8]) for c in readpartitioning)


def test_clusterediting4():
    reads = """
    0010111110111111111001111
    111111111111111111111 111
    011011111011111 111001111
    00101 111011111 1110011 1
     11 11111111 111111111111
    1111111111111111111111 11
    0010111110111111111001111
    111111111111111111111 111
    011011111011111 111001111
    011011111011111 111001111
    """
    # construct a ReadSet
    readset = string_to_readset(reads)

    # compute similarities
    similarities = scoreReadset(readset, 5, 3, 0.06)

    # run cluster editing
    clusterediting = ClusterEditingSolver(similarities, False)
    readpartitioning = clusterediting.run()

    assert any(all(x in c for x in [0, 2, 3, 6, 8, 9]) for c in readpartitioning)
    assert any(all(x in c for x in [1, 4, 5, 7]) for c in readpartitioning)


def test_clusterediting5():
    reads = """
    0010111110111111111001111
    111111111111111111111 111
    011011111011111 111001111
    00101 111011111 1110011 1
     11 11111111 111111111111
    1111111111111111111111 11
    0010111110111111111001111
    111111111111111111111 111
    011011111011111 111001111
    011011111011111 111001111
    """
    # construct a ReadSet
    readset = string_to_readset(reads)

    # compute similarities
    similarities = scoreReadset(readset, 5, 3, 0.01)

    # run cluster editing
    clusterediting = ClusterEditingSolver(similarities, False)
    readpartitioning = clusterediting.run()

    assert any(all(x in c for x in [0, 3, 6]) for c in readpartitioning)
    assert any(all(x in c for x in [1, 4, 5, 7]) for c in readpartitioning)
    assert any(all(x in c for x in [2, 8, 9]) for c in readpartitioning)



"""
Compare two or more phased variant files
"""
import logging
import math
from collections import defaultdict
from contextlib import ExitStack
import dataclasses
from itertools import chain, permutations
from typing import Set, List, Optional, DefaultDict, Dict

from whatshap.vcf import VcfReader, VcfVariant, VariantTable, PloidyError
from whatshap.core import Genotype
from whatshap.polyphase_solver import SwitchFlipCalculator
from whatshap.cli import CommandLineError


logger = logging.getLogger(__name__)

COUNT_WIDTH = 9


# fmt: off
def add_arguments(parser):
    add = parser.add_argument
    add('--sample', metavar='SAMPLE', default=None, help='Name of the sample '
        'to process. If not given, use first sample found in VCF.')
    add('--names', metavar='NAMES', default=None, help='Comma-separated list '
        'of data set names to be used in the report (in same order as VCFs).')
    add('--ignore-sample-name', default=False, action='store_true', help='For single '
        'sample VCFs, ignore sample name and assume all samples are the same.')
    add('--tsv-pairwise', metavar='TSVPAIRWISE', default=None, help='Filename to write '
        'comparison results from pair-wise comparison to (tab-separated).')
    add('--tsv-multiway', metavar='TSVMULTIWAY', default=None, help='Filename to write '
        'comparison results from multiway comparison to (tab-separated). Only for diploid VCFs.')
    add('--only-snvs', default=False, action="store_true", help='Only process SNVs '
        'and ignore all other variants.')
    add('--switch-error-bed', default=None, help='Write BED file with switch error positions '
        'to given filename. Only for diploid VCFs.')
    add('--plot-blocksizes', default=None, help='Write PDF file with a block length histogram '
        'to given filename (requires matplotlib).')
    add('--plot-sum-of-blocksizes', default=None, help='Write PDF file with a block length histogram in which the height of each bar corresponds to the sum of lengths.')
    add('--longest-block-tsv', default=None, help='Write position-wise agreement of longest '
        'joint blocks in each chromosome to tab-separated file. Only for diploid VCFs.')
    add('--ploidy', '-p', metavar='PLOIDY', type=int, default=2, help='The ploidy of the sample(s) (default: %(default)s).')
    # TODO: what's the best way to request "two or more" VCFs?
    add('vcf', nargs='+', metavar='VCF/BCF', help='At least two phased variant files (VCF or BCF) to be compared.')
# fmt: on


def validate(args, parser):
    if len(args.vcf) < 2:
        parser.error("At least two VCFs need to be given.")
    if args.ploidy < 2:
        parser.error("Ploidy must be > 1.")
    if args.ploidy > 2 and args.tsv_multiway:
        parser.error("Option --tsv-multiway can only be used if ploidy=2.")
    if args.ploidy > 2 and args.switch_error_bed:
        parser.error("Option --switch-error-bed can only be used if ploidy=2.")
    if args.ploidy > 2 and args.longest_block_tsv:
        parser.error("Option --longest-block-tsv can only be used if ploidy=2.")


class SwitchFlips:
    def __init__(self, switches: int = 0, flips: int = 0):
        self.switches: int = switches
        self.flips: int = flips

    def __iadd__(self, other):
        self.switches += other.switches
        self.flips += other.flips
        return self

    def __repr__(self):
        return f"SwitchFlips(switches={self.switches}, flips={self.flips})"

    def __str__(self):
        return f"{self.switches}/{self.flips}"


class PhasingErrors:
    def __init__(
        self,
        switches: int = 0,
        hamming: int = 0,
        switch_flips: Optional[SwitchFlips] = None,
        diff_genotypes: int = 0,
    ):
        self.switches = switches
        self.hamming = hamming
        self.switch_flips = SwitchFlips() if switch_flips is None else switch_flips
        self.diff_genotypes = diff_genotypes

    def __iadd__(self, other: object) -> "PhasingErrors":
        if not isinstance(other, PhasingErrors):
            raise TypeError("Can only add to PhasingErrors")
        self.switches += other.switches
        self.hamming += other.hamming
        self.switch_flips += other.switch_flips
        self.diff_genotypes += other.diff_genotypes
        return self

    def __repr__(self):
        return "PhasingErrors(switches={}, hamming={}, switch_flips={}, diff_genotypes={})".format(
            self.switches, self.hamming, self.switch_flips, self.diff_genotypes
        )


def complement(s):
    """
    >>> complement('01100')
    '10011'
    """
    t = {"0": "1", "1": "0"}
    return "".join(t[c] for c in s)


def hamming(s0, s1):
    """
    >>> hamming('ABCD', 'AXCY')
    2
    """
    assert len(s0) == len(s1)
    return sum(c0 != c1 for c0, c1 in zip(s0, s1))


def switch_encoding(phasing):
    """
    >>> switch_encoding('0001011')
    '001110'
    """
    assert isinstance(phasing, str)
    return "".join(("0" if phasing[i - 1] == phasing[i] else "1") for i in range(1, len(phasing)))


def compute_switch_flips(phasing0, phasing1) -> SwitchFlips:
    """
    >>> compute_switch_flips("00011", "00100")
    SwitchFlips(switches=1, flips=0)
    >>> compute_switch_flips("00011", "00111")
    SwitchFlips(switches=0, flips=1)
    >>> compute_switch_flips("000", "001")
    SwitchFlips(switches=1, flips=0)
    """
    assert len(phasing0) == len(phasing1)
    s0 = switch_encoding(phasing0)
    s1 = switch_encoding(phasing1)
    result = SwitchFlips()
    switches_in_a_row = 0
    for i, (p0, p1) in enumerate(zip(s0, s1)):
        if p0 != p1:
            switches_in_a_row += 1
        if (i + 1 == len(s0)) or (p0 == p1):
            result.flips += switches_in_a_row // 2
            result.switches += switches_in_a_row % 2
            switches_in_a_row = 0

    return result


def compute_matching_genotype_pos(phasing0, phasing1):
    """
    Computes the positions on which both phasings agree on the genotype.
    """
    assert len(phasing0) == len(phasing1)
    assert len(phasing0) >= 2
    assert len(phasing0[0]) == len(phasing1[0])
    assert all(len(phasing0[i]) == len(phasing0[0]) for i in range(1, len(phasing0)))
    num_vars = len(phasing0[0])
    matching_pos = [
        i
        for i in range(num_vars)
        if Genotype([int(hap[i]) for hap in phasing0])
        == Genotype([int(hap[i]) for hap in phasing1])
    ]
    return matching_pos


def compute_switch_errors_poly(phasing0, phasing1, matching_pos=None):
    """
    Computes the number of necessary switches to transform phasing 0 into phasing 1 or vice versa.
    Positions with non-matching genotypes are omitted.
    """
    assert len(phasing0) == len(phasing1)
    assert len(phasing0) >= 2
    assert len(phasing0[0]) == len(phasing1[0])
    assert all(len(phasing0[i]) == len(phasing0[0]) for i in range(1, len(phasing0)))
    num_vars = len(phasing0[0])

    # If positions with matching genotypes are not precomputed, do it here!
    if matching_pos is None:
        matching_pos = compute_matching_genotype_pos(phasing0, phasing1)

    phasing0_matched = ["".join([hap[i] for i in matching_pos]) for hap in phasing0]
    phasing1_matched = ["".join([hap[i] for i in matching_pos]) for hap in phasing1]

    vector_error = compute_switch_flips_poly(
        phasing0_matched,
        phasing1_matched,
        switch_cost=1,
        flip_cost=2 * num_vars * len(phasing0) + 1,
    )
    assert vector_error.flips == 0

    return vector_error.switches


def compute_switch_flips_poly(phasing0, phasing1, switch_cost=1, flip_cost=1):
    """
    Computes the combined number of switches and flips, which are needed to transform phasing 0 into
    phasing 1 or vice versa.
    """
    (result, switches_in_column, flips_in_column, poswise_config) = compute_switch_flips_poly_bt(
        phasing0, phasing1, switch_cost=switch_cost, flip_cost=flip_cost
    )
    return result


def compute_switch_flips_poly_bt(
    phasing0, phasing1, report_error_positions=False, switch_cost=1, flip_cost=1
):
    # Check input
    if len(phasing0) != len(phasing1):
        logger.error(
            "Incompatible phasings. Number of haplotypes is not equal "
            f"({len(phasing0)} != {len(phasing1)})."
        )
    assert len(phasing0) == len(phasing1)

    num_pos = len(phasing0[0])
    if num_pos == 0:
        return SwitchFlips(), None, None, None
    ploidy = len(phasing0)
    if ploidy == 0:
        return SwitchFlips(), None, None, None
    for i in range(0, len(phasing1)):
        if len(phasing1[i]) != num_pos:
            logger.error(
                "Inconsistent input for phasing. Haplotypes have different lengths "
                f"( len(phasing1[0]={num_pos} != len(phasing1[{i}]={len(phasing1[i])}."
            )
        assert len(phasing1[i]) == num_pos
        if len(phasing0[i]) != num_pos:
            logger.error(
                "Inconsistent input for phasing. Haplotypes have different lengths "
                f"( len(phasing1[0]={num_pos} != len(phasing0[{i}]={len(phasing0[i])}."
            )
        assert len(phasing1[i]) == num_pos
    if ploidy > 6:
        logger.warning(
            "Computing vector error with more than 6 haplotypes. This may take very long ..."
        )

    # Compute comparison
    calc = SwitchFlipCalculator(ploidy, switch_cost, flip_cost)
    result = SwitchFlips()
    (
        switches,
        flips,
        switches_in_column,
        flips_in_column,
        positionwise_config,
    ) = calc.compute_switch_flips_poly(phasing0, phasing1)

    # Aggregate results
    result.switches = switches / ploidy
    result.flips = flips / ploidy
    return result, switches_in_column, flips_in_column, positionwise_config


def poly_num_switches(perm0, perm1):
    cost = 0
    for i in range(len(perm0)):
        if perm0[i] != perm1[i]:
            cost += 1
    return cost


def compare_block(phasing0, phasing1):
    """Input are two lists of haplotype sequences over {0,1}."""
    assert len(phasing0) == len(phasing1)
    ploidy = len(phasing0)

    minimum_hamming_distance = float("inf")
    # compute minimum hamming distance
    for permutation in permutations(phasing0):
        # compute sum of hamming distances
        total_hamming = 0
        for i in range(ploidy):
            total_hamming += hamming(phasing1[i], permutation[i])
        total_hamming /= float(ploidy)
        minimum_hamming_distance = min(minimum_hamming_distance, total_hamming)

    matching_pos = compute_matching_genotype_pos(phasing0, phasing1)

    if ploidy == 2:
        # conversion to int is allowed, as there should be no fractional error counts for diploid comparisons
        switches = int(hamming(switch_encoding(phasing0[0]), switch_encoding(phasing1[0])))
        switch_flips = compute_switch_flips(phasing0[0], phasing1[0])
        minimum_hamming_distance = int(minimum_hamming_distance)
    else:
        switches = compute_switch_errors_poly(phasing0, phasing1, matching_pos)
        switch_flips = compute_switch_flips_poly(phasing0, phasing1)

    return PhasingErrors(
        switches=switches,
        hamming=minimum_hamming_distance,
        switch_flips=switch_flips,
        diff_genotypes=len(phasing0[0]) - len(matching_pos),
    )


def fraction2percentstr(nominator, denominator):
    if denominator == 0:
        return "--"
    else:
        return f"{nominator * 100.0 / denominator:.2f}%"


def safefraction(nominator, denominator):
    if denominator == 0:
        return float("nan")
    else:
        return nominator / denominator


class BedCreator:
    def __init__(self, chromosome: str, dataset_names: List[str]):
        self._chromosome = chromosome
        self._annotation = "{}<-->{}".format(*dataset_names)

    def records(self, phasing0, phasing1, positions):
        """
        Determine positions of switch errors between two phasings
        and yield one BED record per switch error (encoded as a tuple).
        The annotation_string is added to each record.
        """
        assert len(phasing0) == len(phasing1) == len(positions)
        switch_encoding0 = switch_encoding(phasing0)
        switch_encoding1 = switch_encoding(phasing1)
        for i, (sw0, sw1) in enumerate(zip(switch_encoding0, switch_encoding1)):
            if sw0 != sw1:
                yield (self._chromosome, positions[i] + 1, positions[i + 1] + 1, self._annotation)


def print_stat(text: str, value=None, value2=None, text_width=37):
    """
    Print a line like this:

         text: value
    """
    text = text.rjust(text_width)
    if value is None:
        assert value2 is None
        print(text)
    else:
        if value == "-":
            value = "-" * COUNT_WIDTH
        else:
            value = str(value).rjust(COUNT_WIDTH)
        if value2 is None:
            print(text + ":", value)
        else:
            print(text + ":", value, str(value2).rjust(COUNT_WIDTH))


def print_errors(errors, phased_pairs):
    print_stat("phased pairs of variants assessed", phased_pairs)
    print_stat("switch errors", errors.switches)
    print_stat("switch error rate", fraction2percentstr(errors.switches, phased_pairs))
    print_stat("switch/flip decomposition", errors.switch_flips)
    print_stat(
        "switch/flip rate",
        fraction2percentstr(errors.switch_flips.switches + errors.switch_flips.flips, phased_pairs),
    )


@dataclasses.dataclass
class PairwiseComparisonResults:
    intersection_blocks: int
    covered_variants: int
    all_assessed_pairs: int
    all_switches: int
    all_switch_rate: float
    all_switchflips: SwitchFlips
    all_switchflip_rate: float
    blockwise_hamming: int
    blockwise_hamming_rate: int
    blockwise_diff_genotypes: int
    blockwise_diff_genotypes_rate: int
    largestblock_assessed_pairs: int
    largestblock_switches: int
    largestblock_switch_rate: float
    largestblock_switchflips: SwitchFlips
    largestblock_switchflip_rate: float
    largestblock_hamming: int
    largestblock_hamming_rate: float
    largestblock_diff_genotypes: int
    largestblock_diff_genotypes_rate: float


@dataclasses.dataclass
class BlockStats:
    variant_count: int
    span: int


def collect_common_variants(
    variant_tables: List[VariantTable], sample_names: List[str]
) -> Set[VcfVariant]:
    common_variants = None
    for variant_table, sample in zip(variant_tables, sample_names):
        het_variants = [
            v
            for v, gt in zip(variant_table.variants, variant_table.genotypes_of(sample))
            if not gt.is_homozygous()
        ]
        if common_variants is None:
            common_variants = set(het_variants)
        else:
            common_variants.intersection_update(het_variants)
    assert common_variants is not None
    return common_variants


def compare(
    variant_tables: List[VariantTable],
    sample_names: List[str],
    dataset_names: List[str],
    ploidy: int,
):
    """
    Return a PairwiseComparisonResults object if the variant_tables has a length of 2.
    """
    assert len(variant_tables) > 1

    common_variants = collect_common_variants(variant_tables, sample_names)
    assert common_variants is not None

    print_stat("common heterozygous variants", len(common_variants))
    print_stat("(restricting to these below)")
    phases = []
    sorted_variants = sorted(common_variants, key=lambda v: v.position)
    for variant_table, sample in zip(variant_tables, sample_names):
        p = [
            phase
            for variant, phase in zip(variant_table.variants, variant_table.phases_of(sample))
            if variant in common_variants
        ]
        assert [v for v in variant_table.variants if v in common_variants] == sorted_variants
        assert len(p) == len(common_variants)
        phases.append(p)

    # blocks[variant_table_index][block_id] is a list of indices into common_variants
    blocks: List[DefaultDict[int, List[int]]] = [defaultdict(list) for _ in variant_tables]
    block_intersection = defaultdict(list)
    for variant_index in range(len(common_variants)):
        any_none = False
        for i in range(len(phases)):
            phase = phases[i][variant_index]
            if phase is None or any(p is None for p in phase.phase):
                any_none = True
            else:
                blocks[i][phase.block_id].append(variant_index)
        if not any_none:
            joint_block_id = tuple(
                phase[variant_index].block_id for phase in phases  # type: ignore
            )
            block_intersection[joint_block_id].append(variant_index)

    # create statistics on each block in each data set
    block_stats = compute_block_stats(blocks, sorted_variants)

    for dataset_name, blck in zip(dataset_names, blocks):
        print_stat(
            f"non-singleton blocks in {dataset_name}",
            len([b for b in blck.values() if len(b) > 1]),
        )
        print_stat("--> covered variants", sum(len(b) for b in blck.values() if len(b) > 1))

    intersection_block_count = sum(1 for b in block_intersection.values() if len(b) > 1)
    intersection_block_variants = sum(len(b) for b in block_intersection.values() if len(b) > 1)
    print_stat("non-singleton intersection blocks", intersection_block_count)
    print_stat("--> covered variants", intersection_block_variants)
    if len(variant_tables) == 2:
        (
            bed_records,
            longest_block_agreement,
            longest_block_positions,
            pairwise_comparison,
        ) = compare_pair(
            block_intersection,
            intersection_block_count,
            intersection_block_variants,
            phases,
            ploidy,
            sorted_variants,
            BedCreator(variant_tables[0].chromosome, dataset_names),
        )

        return (
            pairwise_comparison,
            bed_records,
            block_stats,
            longest_block_positions,
            longest_block_agreement,
            None,
        )
    else:
        assert ploidy == 2
        multiway_results = compare_multiway(block_intersection, dataset_names, phases)
        return None, None, block_stats, None, None, multiway_results


def compare_pair(
    block_intersection,
    intersection_block_count,
    intersection_block_variants,
    phases,
    ploidy,
    sorted_variants,
    bed_creator: Optional[BedCreator],
):
    longest_block = 0
    longest_block_errors = PhasingErrors()
    longest_block_positions = []
    longest_block_agreement = []
    phased_pairs = 0
    bed_records = []
    total_errors = PhasingErrors()
    total_compared_variants = 0
    for block in block_intersection.values():
        if len(block) < 2:
            continue
        phasing0 = []
        phasing1 = []
        for j in range(ploidy):
            p0 = "".join(str(phases[0][i].phase[j]) for i in block)
            p1 = "".join(str(phases[1][i].phase[j]) for i in block)
            phasing0.append(p0)
            phasing1.append(p1)
        block_positions = [sorted_variants[i].position for i in block]
        errors = compare_block(phasing0, phasing1)

        # TODO: extend to polyploid
        if ploidy == 2 and bed_creator is not None:
            bed_records.extend(bed_creator.records(phasing0[0], phasing1[0], block_positions))
        total_errors += errors
        phased_pairs += len(block) - 1
        total_compared_variants += len(block)
        if len(block) > longest_block:
            longest_block = len(block)
            longest_block_errors = errors
            longest_block_positions = block_positions
            # TODO: extend to polyploid
            if ploidy == 2:
                if hamming(phasing0, phasing1) < hamming(phasing0[0], complement(phasing1[0])):
                    longest_block_agreement = [
                        1 * (p0 == p1) for p0, p1 in zip(phasing0[0], phasing1[0])
                    ]
                else:
                    longest_block_agreement = [
                        1 * (p0 != p1) for p0, p1 in zip(phasing0[0], phasing1[0])
                    ]
    longest_block_assessed_pairs = max(longest_block - 1, 0)
    print_stat("ALL INTERSECTION BLOCKS", "-")
    print_errors(total_errors, phased_pairs)
    print_stat("Block-wise Hamming distance", total_errors.hamming)
    print_stat(
        "Block-wise Hamming distance [%]",
        fraction2percentstr(total_errors.hamming, total_compared_variants),
    )
    print_stat("Different genotypes", total_errors.diff_genotypes)
    print_stat(
        "Different genotypes [%]",
        fraction2percentstr(total_errors.diff_genotypes, total_compared_variants),
    )
    print_stat("LARGEST INTERSECTION BLOCK", "-")
    print_errors(longest_block_errors, longest_block_assessed_pairs)
    print_stat("Hamming distance", longest_block_errors.hamming)
    print_stat(
        "Hamming distance [%]", fraction2percentstr(longest_block_errors.hamming, longest_block)
    )
    print_stat("Different genotypes", longest_block_errors.diff_genotypes)
    print_stat(
        "Different genotypes [%]",
        fraction2percentstr(longest_block_errors.diff_genotypes, longest_block),
    )
    pcr = PairwiseComparisonResults(
        intersection_blocks=intersection_block_count,
        covered_variants=intersection_block_variants,
        all_assessed_pairs=phased_pairs,
        all_switches=total_errors.switches,
        all_switch_rate=safefraction(total_errors.switches, phased_pairs),
        all_switchflips=total_errors.switch_flips,
        all_switchflip_rate=safefraction(
            total_errors.switch_flips.switches + total_errors.switch_flips.flips, phased_pairs
        ),
        blockwise_hamming=total_errors.hamming,
        blockwise_hamming_rate=safefraction(total_errors.hamming, total_compared_variants),
        blockwise_diff_genotypes=total_errors.diff_genotypes,
        blockwise_diff_genotypes_rate=safefraction(
            total_errors.diff_genotypes, total_compared_variants
        ),
        largestblock_assessed_pairs=longest_block_assessed_pairs,
        largestblock_switches=longest_block_errors.switches,
        largestblock_switch_rate=safefraction(
            longest_block_errors.switches, longest_block_assessed_pairs
        ),
        largestblock_switchflips=longest_block_errors.switch_flips,
        largestblock_switchflip_rate=safefraction(
            longest_block_errors.switch_flips.switches + longest_block_errors.switch_flips.flips,
            longest_block_assessed_pairs,
        ),
        largestblock_hamming=longest_block_errors.hamming,
        largestblock_hamming_rate=safefraction(longest_block_errors.hamming, longest_block),
        largestblock_diff_genotypes=longest_block_errors.diff_genotypes,
        largestblock_diff_genotypes_rate=safefraction(
            longest_block_errors.diff_genotypes, longest_block
        ),
    )
    return bed_records, longest_block_agreement, longest_block_positions, pcr


def compare_multiway(block_intersection, dataset_names, phases):
    histogram = defaultdict(int)
    total_compared = 0
    for block in block_intersection.values():
        if len(block) < 2:
            continue
        total_compared += len(block) - 1
        phasings = ["".join(str(phases[j][i].phase[0]) for i in block) for j in range(len(phases))]
        switch_encodings = [switch_encoding(p) for p in phasings]
        for i in range(len(block) - 1):
            s = "".join(switch_encodings[j][i] for j in range(len(switch_encodings)))
            s = min(s, complement(s))
            histogram[s] += 1
    print_stat("Compared pairs of variants", total_compared)
    bipartitions = list(histogram.keys())
    bipartitions.sort()
    multiway_results = {}  # (dataset_list0, dataset_list1) --> count
    for i, s in enumerate(bipartitions):
        count = histogram[s]
        if i == 0:
            assert {c for c in s} == set("0")
            print("ALL AGREE")
        elif i == 1:
            print("DISAGREEMENT")
        left, right = [], []
        for name, leftright in zip(dataset_names, s):
            if leftright == "0":
                left.append(name)
            else:
                right.append(name)
        print_stat(
            ("{{{}}} vs. {{{}}}".format(",".join(left), ",".join(right))),
            count,
            fraction2percentstr(count, total_compared),
        )
        multiway_results[(",".join(left), ",".join(right))] = count
    return multiway_results


def compute_block_stats(
    blocks: List[DefaultDict[int, List[int]]], sorted_variants: List[VcfVariant]
):
    block_stats = []
    for block in blocks:
        l = []
        for block_id, variant_indices in block.items():
            if len(variant_indices) < 2:
                continue
            span = (
                sorted_variants[variant_indices[-1]].position
                - sorted_variants[variant_indices[0]].position
            )
            l.append(BlockStats(len(variant_indices), span))
        block_stats.append(l)
    return block_stats


def create_blocksize_histogram(filename, block_stats, names, use_weights=False):
    try:
        import matplotlib
        import numpy

        matplotlib.use("pdf")
        from matplotlib import pyplot
        from matplotlib.backends.backend_pdf import PdfPages
    except ImportError:
        raise CommandLineError(
            "To use option --plot-blocksizes, you need to have numpy and matplotlib installed."
        )

    assert len(block_stats) == len(names)

    color_list = ["#ffa347", "#0064c8", "#b42222", "#22a5b4", "#b47c22", "#6db6ff"]
    if len(color_list) < len(block_stats):
        color_count = len(block_stats)
        color_list = pyplot.cm.Set1([n / color_count for n in range(color_count)])
    colors = color_list[: len(block_stats)]

    with PdfPages(filename) as pdf:
        for what, xlabel in [
            (lambda stats: stats.variant_count, "variant count"),
            (lambda stats: stats.span, "span [bp]"),
        ]:
            pyplot.figure(figsize=(10, 8))
            max_value = max(what(stats) for stats in chain(*block_stats))
            common_bins = numpy.logspace(0, math.ceil(math.log10(max_value)), 50)
            for l, name, color in zip(block_stats, names, colors):
                x = [what(stats) for stats in l]
                n, bins, patches = pyplot.hist(
                    x,
                    bins=common_bins,
                    alpha=0.6,
                    color=color,
                    label=name,
                    weights=x if use_weights else None,
                )
            pyplot.xlabel(xlabel)
            pyplot.ylabel("Number of blocks")
            pyplot.gca().set_xscale("log")
            pyplot.gca().set_yscale("log")
            pyplot.grid(True)
            pyplot.legend()
            pdf.savefig()
            pyplot.close()

            pyplot.figure(figsize=(10, 8))
            common_bins = numpy.logspace(0, math.ceil(math.log10(max_value)), 25)
            x = [[what(stats) for stats in l] for l in block_stats]
            n, bins, patches = pyplot.hist(
                x,
                bins=common_bins,
                alpha=0.6,
                color=colors,
                label=names,
                weights=x if use_weights else None,
            )
            pyplot.xlabel(xlabel)
            pyplot.ylabel("Number of blocks")
            pyplot.gca().set_xscale("log")
            pyplot.gca().set_yscale("log")
            pyplot.grid(True)
            pyplot.legend()
            pdf.savefig()
            pyplot.close()


def run_compare(
    vcf,
    ploidy,
    names=None,
    sample=None,
    ignore_sample_name=False,
    tsv_pairwise=None,
    tsv_multiway=None,
    only_snvs=False,
    switch_error_bed=None,
    plot_blocksizes=None,
    plot_sum_of_blocksizes=None,
    longest_block_tsv=None,
):
    vcf_readers = [VcfReader(f, indels=not only_snvs, phases=True, ploidy=ploidy) for f in vcf]
    if names:
        dataset_names = names.split(",")
        if len(dataset_names) != len(vcf):
            raise CommandLineError(
                "Number of names given with --names does not equal number of VCFs."
            )
    else:
        dataset_names = [f"file{i}" for i in range(len(vcf))]

    sample_names = get_sample_names(
        vcf_readers, requested_sample=sample, ignore_name=ignore_sample_name
    )

    with ExitStack() as stack:
        tsv_pairwise_file = tsv_multiway_file = longest_block_tsv_file = switch_error_bedfile = None
        if tsv_pairwise:
            tsv_pairwise_file = stack.enter_context(open(tsv_pairwise, "w"))

        if tsv_multiway:
            tsv_multiway_file = stack.enter_context(open(tsv_multiway, "w"))
            print(
                "#sample",
                "chromosome",
                "dataset_list0",
                "dataset_list1",
                "count",
                sep="\t",
                file=tsv_multiway_file,
            )

        if longest_block_tsv:
            longest_block_tsv_file = stack.enter_context(open(longest_block_tsv, "w"))
            print(
                "#dataset_name0",
                "dataset_name1",
                "#sample",
                "chromosome",
                "position",
                "phase_agreeing",
                sep="\t",
                file=longest_block_tsv_file,
            )

        if tsv_pairwise_file:
            fields = [
                "#sample",
                "chromosome",
                "dataset_name0",
                "dataset_name1",
                "file_name0",
                "file_name1",
            ]
            field_names = [f.name for f in dataclasses.fields(PairwiseComparisonResults)]
            fields.extend(field_names)
            fields.extend(["het_variants0", "only_snvs"])
            print(*fields, sep="\t", file=tsv_pairwise_file)

        if switch_error_bed:
            switch_error_bedfile = stack.enter_context(open(switch_error_bed, "w"))

        if len(set(sample_names)) > 1 and ignore_sample_name:
            print(
                "Comparing phasings for samples:",
                ", ".join(sample_names),
                " (--ignore-sample-names selected)",
            )
        else:
            print("Comparing phasings for sample", sample_names[0])

        vcfs = get_variant_tables(vcf_readers, vcf)
        chromosomes = get_common_chromosomes(vcfs)
        if len(chromosomes) == 0:
            raise CommandLineError("No chromosome is contained in all VCFs. Aborting.")
        logger.info("Chromosomes present in all VCFs: %s", ", ".join(chromosomes))

        print("FILENAMES")
        longest_name = max(len(n) for n in dataset_names)
        for name, filename in zip(dataset_names, vcf):
            print(name.rjust(longest_name + 2), "=", filename)

        width = max(longest_name, 15) + 5

        all_block_stats = [[] for _ in vcfs]

        def add_block_stats(block_stats):
            assert len(block_stats) == len(all_block_stats)
            for big_list, new_list in zip(all_block_stats, block_stats):
                big_list.extend(new_list)

        for chromosome in sorted(chromosomes):
            print(f"---------------- Chromosome {chromosome} ----------------")
            all_bed_records = []
            variant_tables = [vcf[chromosome] for vcf in vcfs]
            all_variants_union = set()
            all_variants_intersection = None
            het_variants_union = set()
            het_variants_intersection = None
            het_variant_sets = []
            het_variants0 = None
            print("VARIANT COUNTS (heterozygous / all): ")
            for variant_table, name, sample in zip(variant_tables, dataset_names, sample_names):
                all_variants_union.update(variant_table.variants)
                het_variants = [
                    v
                    for v, gt in zip(variant_table.variants, variant_table.genotypes_of(sample))
                    if not gt.is_homozygous()
                ]
                if het_variants0 is None:
                    het_variants0 = len(het_variants)
                het_variants_union.update(het_variants)
                if all_variants_intersection is None:
                    all_variants_intersection = set(variant_table.variants)
                    het_variants_intersection = set(het_variants)
                else:
                    all_variants_intersection.intersection_update(variant_table.variants)
                    het_variants_intersection.intersection_update(het_variants)
                het_variant_sets.append(set(het_variants))
                print(
                    f"{name}:".rjust(width),
                    str(len(het_variants)).rjust(COUNT_WIDTH),
                    "/",
                    str(len(variant_table.variants)).rjust(COUNT_WIDTH),
                )
            print(
                "UNION:".rjust(width),
                str(len(het_variants_union)).rjust(COUNT_WIDTH),
                "/",
                str(len(all_variants_union)).rjust(COUNT_WIDTH),
            )
            print(
                "INTERSECTION:".rjust(width),
                str(len(het_variants_intersection)).rjust(COUNT_WIDTH),
                "/",
                str(len(all_variants_intersection)).rjust(COUNT_WIDTH),
            )

            for i in range(len(vcfs)):
                for j in range(i + 1, len(vcfs)):
                    print(
                        "PAIRWISE COMPARISON: {} <--> {}:".format(
                            dataset_names[i], dataset_names[j]
                        )
                    )
                    (
                        results,
                        bed_records,
                        block_stats,
                        longest_block_positions,
                        longest_block_agreement,
                        multiway_results,
                    ) = compare(
                        [variant_tables[i], variant_tables[j]],
                        [sample_names[i], sample_names[j]],
                        [dataset_names[i], dataset_names[j]],
                        ploidy,
                    )
                    if len(vcfs) == 2:
                        add_block_stats(block_stats)
                    all_bed_records.extend(bed_records)
                    sample_name = (
                        f"{sample_names[i]}_{sample_names[j]}"
                        if ignore_sample_name
                        else sample_names[i]
                    )
                    if tsv_pairwise_file:
                        fields = [
                            sample_name,
                            chromosome,
                            dataset_names[i],
                            dataset_names[j],
                            vcf[i],
                            vcf[j],
                        ]
                        fields.extend(dataclasses.astuple(results))
                        fields.extend([het_variants0, int(only_snvs)])
                        print(*fields, sep="\t", file=tsv_pairwise_file)
                    if longest_block_tsv_file:
                        assert ploidy == 2
                        assert len(longest_block_positions) == len(longest_block_agreement)
                        for position, phase_agreeing in zip(
                            longest_block_positions, longest_block_agreement
                        ):
                            print(
                                dataset_names[i],
                                dataset_names[j],
                                sample_name,
                                chromosome,
                                position,
                                phase_agreeing,
                                sep="\t",
                                file=longest_block_tsv_file,
                            )

            # if requested, write all switch errors found in the current chromosome to the bed file
            if switch_error_bedfile:
                assert ploidy == 2
                all_bed_records.sort()
                for record in all_bed_records:
                    print(*record, sep="\t", file=switch_error_bedfile)

            if len(vcfs) > 2:
                assert ploidy == 2
                print("MULTIWAY COMPARISON OF ALL PHASINGS:")
                (
                    results,
                    bed_records,
                    block_stats,
                    longest_block_positions,
                    longest_block_agreement,
                    multiway_results,
                ) = compare(variant_tables, sample_names, dataset_names, ploidy)
                add_block_stats(block_stats)
                if tsv_multiway_file:
                    sample_name = (
                        "_".join(set(sample_names)) if ignore_sample_name else sample_names[0]
                    )
                    for ((dataset_list0, dataset_list1), count) in multiway_results.items():
                        print(
                            sample_name,
                            chromosome,
                            "{" + dataset_list0 + "}",
                            "{" + dataset_list1 + "}",
                            count,
                            sep="\t",
                            file=tsv_multiway_file,
                        )

        if plot_blocksizes:
            create_blocksize_histogram(plot_blocksizes, all_block_stats, dataset_names)
        if plot_sum_of_blocksizes:
            create_blocksize_histogram(
                plot_sum_of_blocksizes, all_block_stats, dataset_names, use_weights=True
            )


def get_common_chromosomes(vcfs: List[Dict[str, VariantTable]]) -> List[str]:
    common = None
    for chrom_variant_table_map in vcfs:
        chromosomes = chrom_variant_table_map.keys()
        if common is None:
            common = set(chromosomes)
        else:
            common.intersection_update(chromosomes)
    if common is None:
        return []
    return sorted(common)


def get_variant_tables(
    vcf_readers: List[VcfReader], vcf_filenames: List[str]
) -> List[Dict[str, VariantTable]]:
    vcfs = []
    for reader, filename in zip(vcf_readers, vcf_filenames):
        # create dict mapping chromosome names to VariantTables
        m = dict()
        logger.info("Reading phasing from %r", filename)
        try:
            for variant_table in reader:
                m[variant_table.chromosome] = variant_table
        except PloidyError as e:
            raise CommandLineError(f"Provided ploidy is invalid: {e}. Aborting.")
        vcfs.append(m)
    return vcfs


def get_sample_names(
    vcf_readers: List[VcfReader], requested_sample: Optional[str], ignore_name: bool = False
) -> List[str]:
    first_samples = []
    sample_intersection = None
    for vcf_reader in vcf_readers:
        if sample_intersection is None:
            sample_intersection = set(vcf_reader.samples)
        else:
            sample_intersection.intersection_update(vcf_reader.samples)

        if ignore_name and len(vcf_reader.samples) > 1:
            raise CommandLineError(
                "File '{file}' contains multiple samples, option --ignore-sample-name not available.".format(
                    file=vcf_reader.path
                )
            )
        first_samples.append(vcf_reader.samples[0])
    assert sample_intersection is not None
    if requested_sample:
        sample_intersection.intersection_update([requested_sample])
        if len(sample_intersection) == 0:
            raise CommandLineError(
                "Sample {!r} requested on command-line not found in all VCFs".format(
                    requested_sample
                )
            )
        sample_names = [requested_sample] * len(vcf_readers)
    elif ignore_name:
        sample_names = first_samples
    else:
        if len(sample_intersection) == 0:
            raise CommandLineError("None of the samples is present in all VCFs")
        elif len(sample_intersection) == 1:
            sample_names = [list(sample_intersection)[0]] * len(vcf_readers)
        else:
            raise CommandLineError(
                "More than one sample is present in all VCFs, please use"
                " --sample to specify which sample to work on."
            )
    return sample_names


def main(args):
    run_compare(**vars(args))



from whatshap.priorityqueue import PriorityQueue


def test_queue():
    pq = PriorityQueue()
    pq.push(10, ord("a"))
    pq.push(5, ord("b"))
    pq.push(12, ord("c"))
    pq.push(3, ord("d"))
    assert len(pq) == 4
    assert pq.pop() == (12, ord("c"))
    assert pq.pop() == (10, ord("a"))
    assert pq.pop() == (5, ord("b"))
    assert pq.pop() == (3, ord("d"))


def test_queue2():
    pq = PriorityQueue()
    pq.push(1, ord("a"))
    pq.push(50, ord("b"))
    pq.push(2, ord("c"))
    pq.push(30, ord("d"))
    assert len(pq) == 4
    assert pq.get_score_by_item(ord("d")) == 30
    assert pq.pop() == (50, ord("b"))
    assert pq.pop() == (30, ord("d"))
    assert pq.pop() == (2, ord("c"))
    assert pq.pop() == (1, ord("a"))


def test_change_score():
    pq = PriorityQueue()
    pq.push(10, ord("a"))
    pq.push(5, ord("b"))
    pq.change_score(ord("a"), 2)
    pq.push(12, ord("c"))
    pq.push(3, ord("d"))
    pq.change_score(ord("c"), 1)
    pq.change_score(ord("d"), 15)
    assert len(pq) == 4
    assert pq.pop() == (15, ord("d"))
    assert pq.pop() == (5, ord("b"))
    assert pq.pop() == (2, ord("a"))
    assert pq.pop() == (1, ord("c"))


def test_change_score_sorting():
    pq = PriorityQueue()
    pq.push(50, ord("1"))
    pq.push(40, ord("2"))
    pq.push(30, ord("3"))
    pq.push(20, ord("4"))
    pq.push(10, ord("5"))
    pq.change_score(ord("5"), 100)
    pq.change_score(ord("2"), 45)
    assert len(pq) == 5
    assert pq.pop() == (100, ord("5"))
    # assert pq.pop() == (45,ord('2'))
    pq.push(60, ord("8"))
    assert pq.pop() == (60, ord("8"))
    pq.change_score(ord("2"), 40)
    assert pq.pop() == (50, ord("1"))
    assert pq.pop() == (40, ord("2"))


def test_is_empty():
    pq = PriorityQueue()
    assert pq.is_empty()
    pq.push(10, ord("A"))
    assert not pq.is_empty()
    pq.pop()
    assert pq.is_empty()
    pq.push(9, ord("B"))
    assert not pq.is_empty()
    pq.push(8, ord("C"))
    assert not pq.is_empty()
    pq.pop()
    assert not pq.is_empty()
    pq.pop()
    assert pq.is_empty()
    pq.push(7, ord("D"))
    assert not pq.is_empty()
    pq.push(6, ord("E"))
    assert not pq.is_empty()
    pq.push(5, ord("F"))
    assert not pq.is_empty()
    pq.push(4, ord("G"))
    assert not pq.is_empty()
    pq.pop()
    assert not pq.is_empty()
    pq.pop()
    assert not pq.is_empty()
    pq.pop()
    assert not pq.is_empty()
    pq.pop()
    assert pq.is_empty()


def test_tuple_score():
    pq = PriorityQueue()
    assert pq.is_empty()
    pq.push((4, 0, 2), ord("A"))
    (score, item) = pq.pop()
    assert score == (4, 0, 2)
    assert item == ord("A")
    pq.push((1, 0, 2), ord("B"))
    pq.push((2, 0, 2), ord("C"))
    pq.push((3, 0, 2), ord("D"))
    pq.push((40, 0, 2), ord("E"))
    pq.push((0, 0, 2), ord("F"))
    pq.push((50, 0, 2), ord("G"))
    (score, item) = pq.pop()
    assert score == (50, 0, 2)
    assert pq.get_score_by_item(ord("E")) == (40, 0, 2)
    assert pq.get_score_by_item(ord("F")) == (0, 0, 2)
    assert pq.get_score_by_item(ord("D")) == (3, 0, 2)
    assert item == ord("G")
    (score, item) = pq.pop()
    assert score == (40, 0, 2)
    assert item == ord("E")
    (score, item) = pq.pop()
    assert score == (3, 0, 2)
    assert item == ord("D")
    (score, item) = pq.pop()
    assert score == (2, 0, 2)
    assert item == ord("C")
    pq.pop()
    pq.pop()
    assert pq.is_empty()


def test_tuple_score_sorting():
    pq = PriorityQueue()
    pq.push((10, 0, 0), ord("B"))
    pq.push((10, 2, 6), ord("C"))
    pq.push((10, 3, 2), ord("D"))
    pq.push((10, 4, 3), ord("E"))
    pq.push((10, 2, 2), ord("F"))
    pq.push((10, 0, 2), ord("G"))
    (score, item) = pq.pop()
    assert score == (10, 4, 3)
    assert item == ord("E")
    (score, item) = pq.pop()
    assert score == (10, 3, 2)
    assert item == ord("D")
    (score, item) = pq.pop()
    assert score == (10, 2, 6)
    assert item == ord("C")
    (score, item) = pq.pop()
    assert score == (10, 2, 2)
    assert item == ord("F")
    (score, item) = pq.pop()
    assert score == (10, 0, 2)
    assert item == ord("G")
    pq.push((1, 10, 4), ord("X"))
    pq.push((5, 0, 6), ord("Y"))
    pq.push((1, 8, 2), ord("Z"))
    pq.change_score(ord("Y"), (100, 100, 100))
    pq.change_score(ord("Z"), (0, 0, 0))
    (score, item) = pq.pop()
    assert score == (100, 100, 100)
    assert item == ord("Y")
    (score, item) = pq.pop()
    assert score == (10, 0, 0)
    assert item == ord("B")
    (score, item) = pq.pop()
    assert score == (1, 10, 4)
    assert item == ord("X")
    (score, item) = pq.pop()
    assert score == (0, 0, 0)
    assert item == ord("Z")
    assert pq.is_empty()



# cython: language_level=3

def _iterate_cigar(variants, int j, bam_read, cigartuples):
	"""
	Iterate over the CIGAR of the given bam_read and variants[j:] in lockstep.

	Yield tuples (index, i, consumed, query_pos) where index is into the variants list

	i and consumed describe the split position in the cigar

	bam_read -- a pysam.AlignedSegment
	variants -- list of variants (VcfVariant objects)
	j -- index of the first variant (in the variants list) to check
	"""
	cdef:
		int ref_pos = bam_read.reference_start  # position relative to reference
		int query_pos = 0  # position relative to read
		int cigar_op
		int length
		int i
		int n = len(variants)
		int v_position

	# Skip variants that are located to the left of the read
	while j < n and variants[j].position < ref_pos:
		j += 1

	# Iterate over the CIGAR sequence (defining the alignment) and variant list in lockstep
	for i, (cigar_op, length) in enumerate(cigartuples):
		# The mapping of CIGAR operators to numbers is:
		# MIDNSHPX= => 012345678
		if j < n:
			v_position = variants[j].position
		if cigar_op in (0, 7, 8):  # M, X, = operators (match)
			# Iterate over all variants that are in this matching region
			while j < n and v_position < ref_pos + length:
				assert v_position >= ref_pos
				yield (j, i, v_position - ref_pos, query_pos + v_position - ref_pos)
				j += 1
				if j < n:
					v_position = variants[j].position
			query_pos += length
			ref_pos += length
		elif cigar_op == 1:  # I operator (insertion)
			# TODO it should work to *not* handle the variant here, but at the next M or D region
			if j < n and v_position == ref_pos:
				yield (j, i, 0, query_pos)
				j += 1
				if j < n:
					v_position = variants[j].position
			query_pos += length
		elif cigar_op == 2:  # D operator (deletion)
			# Iterate over all variants that are in this deleted region
			while j < n and v_position < ref_pos + length:
				assert v_position >= ref_pos
				yield (j, i, v_position - ref_pos, query_pos)
				j += 1
				if j < n:
					v_position = variants[j].position
			ref_pos += length
		elif cigar_op == 3:  # N operator (reference skip)
			# Iterate over all variants that are in this skipped region
			while j < n and v_position < ref_pos + length:
				assert v_position >= ref_pos
				j += 1
				if j < n:
					v_position = variants[j].position
			ref_pos += length
		elif cigar_op == 4:  # S operator (soft clipping)
			query_pos += length
		elif cigar_op == 5 or cigar_op == 6:  # H or P (hard clipping or padding)
			pass
		else:
			raise ValueError("Unsupported CIGAR operation: {}".format(cigar_op))



"""
Tests for 'whatshap compare'
"""

from collections import namedtuple
from whatshap.cli.compare import run_compare, compute_switch_flips_poly, compare_block


def test_compare1(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_compare(
        vcf=["tests/data/phased1.vcf", "tests/data/phased2.vcf"],
        ploidy=2,
        names="p1,p2",
        tsv_pairwise=outtsv,
        sample="sample1",
    )
    lines = [l.split("\t") for l in open(outtsv)]
    assert len(lines) == 3
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chrA, entry_chrB = [Fields(*l) for l in lines[1:]]

    assert entry_chrA.chromosome == "chrA"
    assert entry_chrA.all_assessed_pairs == "4"
    assert entry_chrA.all_switches == "1"
    assert entry_chrA.all_switchflips == "1/0"
    assert entry_chrA.blockwise_hamming == "1"
    assert entry_chrA.largestblock_assessed_pairs == "2"
    assert entry_chrA.largestblock_switches == "1"
    assert entry_chrA.largestblock_hamming == "1"

    assert entry_chrB.chromosome == "chrB"
    assert entry_chrB.all_assessed_pairs == "1"
    assert entry_chrB.all_switches == "0"
    assert entry_chrB.all_switchflips == "0/0"
    assert entry_chrB.blockwise_hamming == "0"
    assert entry_chrB.largestblock_assessed_pairs == "1"
    assert entry_chrB.largestblock_switches == "0"
    assert entry_chrB.largestblock_hamming == "0"


def test_compare2(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_compare(
        vcf=["tests/data/phased1.vcf", "tests/data/phased2.vcf"],
        ploidy=2,
        names="p1,p2",
        tsv_pairwise=outtsv,
        sample="sample2",
    )
    lines = [l.split("\t") for l in open(outtsv)]
    assert len(lines) == 3
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chrA, entry_chrB = [Fields(*l) for l in lines[1:]]

    assert entry_chrA.chromosome == "chrA"
    assert entry_chrA.all_assessed_pairs == "6"
    assert entry_chrA.all_switches == "2"
    assert entry_chrA.all_switchflips == "0/1"
    assert entry_chrA.blockwise_hamming == "1"
    assert entry_chrA.largestblock_assessed_pairs == "5"
    assert entry_chrA.largestblock_switches == "2"
    assert entry_chrA.largestblock_hamming == "1"

    assert entry_chrB.chromosome == "chrB"
    assert entry_chrB.all_assessed_pairs == "1"
    assert entry_chrB.all_switches == "1"
    assert entry_chrB.all_switchflips == "1/0"
    assert entry_chrB.blockwise_hamming == "1"
    assert entry_chrB.largestblock_assessed_pairs == "1"
    assert entry_chrB.largestblock_switches == "1"
    assert entry_chrB.largestblock_hamming == "1"


def test_compare_polyploid1(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_compare(
        vcf=["tests/data/phased.poly1.vcf", "tests/data/phased.poly2.vcf"],
        ploidy=4,
        names="p1,p2",
        tsv_pairwise=outtsv,
        sample="sample1",
    )
    with open(outtsv) as t:
        lines = [line.split("\t") for line in t]
    assert len(lines) == 3
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chr21, entry_chr22 = [Fields(*l) for l in lines[1:]]

    assert entry_chr21.chromosome == "chr21"
    assert entry_chr21.all_assessed_pairs == "1"
    assert entry_chr21.all_switches == "0.0"
    assert entry_chr21.all_switchflips == "0.0/0.0"
    assert entry_chr21.blockwise_hamming == "0.0"
    assert entry_chr21.blockwise_diff_genotypes == "0"
    assert entry_chr21.largestblock_assessed_pairs == "1"
    assert entry_chr21.largestblock_switches == "0.0"
    assert entry_chr21.largestblock_hamming == "0.0"
    assert entry_chr21.largestblock_diff_genotypes == "0"

    assert entry_chr22.chromosome == "chr22"
    assert entry_chr22.all_assessed_pairs == "6"
    assert entry_chr22.all_switches == "1.0"
    assert entry_chr22.all_switchflips == "0.0/0.5"
    assert entry_chr22.blockwise_hamming == "0.5"
    assert entry_chr22.blockwise_diff_genotypes == "0"
    assert entry_chr22.largestblock_assessed_pairs == "5"
    assert entry_chr22.largestblock_switches == "1.0"
    assert entry_chr22.largestblock_hamming == "0.5"
    assert entry_chr22.largestblock_diff_genotypes == "0"


def test_compare_polyploid2(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_compare(
        vcf=["tests/data/phased.poly1.vcf", "tests/data/phased.poly2.vcf"],
        ploidy=4,
        names="p1,p2",
        tsv_pairwise=outtsv,
        sample="sample2",
    )
    with open(outtsv) as t:
        lines = [line.split("\t") for line in t]
    assert len(lines) == 3
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chr21, entry_chr22 = [Fields(*l) for l in lines[1:]]

    assert entry_chr21.chromosome == "chr21"
    assert entry_chr21.all_assessed_pairs == "3"
    assert entry_chr21.all_switches == "0.5"
    assert entry_chr21.all_switchflips in ["0.5/0.0", "0.0/0.5"]
    assert entry_chr21.blockwise_hamming == "0.5"
    assert entry_chr21.blockwise_diff_genotypes == "0"
    assert entry_chr21.largestblock_assessed_pairs == "3"
    assert entry_chr21.largestblock_switches == "0.5"
    assert entry_chr21.largestblock_switchflips in ["0.5/0.0", "0.0/0.5"]
    assert entry_chr21.largestblock_hamming == "0.5"
    assert entry_chr21.largestblock_diff_genotypes == "0"

    assert entry_chr22.chromosome == "chr22"
    assert entry_chr22.all_assessed_pairs == "5"
    assert entry_chr22.all_switches == "1.0"
    assert entry_chr22.all_switchflips in ["1.0/0.0", "0.5/0.5", "0.0/1.0"]
    assert entry_chr22.blockwise_hamming == "1.0"
    assert entry_chr22.blockwise_diff_genotypes == "0"
    assert entry_chr22.largestblock_assessed_pairs == "3"
    assert entry_chr22.largestblock_switches == "0.5"
    assert entry_chr22.largestblock_switchflips in ["0.5/0.0", "0.0/0.5"]
    assert entry_chr22.largestblock_hamming == "0.5"
    assert entry_chr22.largestblock_diff_genotypes == "0"


def test_compare_polyploid3(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_compare(
        vcf=["tests/data/phased.poly1.vcf", "tests/data/phased.poly3.vcf"],
        ploidy=4,
        names="p1,p2",
        tsv_pairwise=outtsv,
        sample="sample1",
    )
    with open(outtsv) as t:
        lines = [line.split("\t") for line in t]
    assert len(lines) == 3
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chr21, entry_chr22 = [Fields(*l) for l in lines[1:]]
    assert entry_chr21.chromosome == "chr21"
    assert entry_chr21.all_assessed_pairs == "2"
    assert entry_chr21.all_switches == "0.0"
    assert entry_chr21.all_switchflips == "0.0/0.0"
    assert entry_chr21.blockwise_hamming == "0.0"
    assert entry_chr21.blockwise_diff_genotypes == "0"
    assert entry_chr21.largestblock_assessed_pairs == "2"
    assert entry_chr21.largestblock_switches == "0.0"
    assert entry_chr21.largestblock_switchflips == "0.0/0.0"
    assert entry_chr21.largestblock_hamming == "0.0"
    assert entry_chr21.largestblock_diff_genotypes == "0"

    assert entry_chr22.chromosome == "chr22"
    assert entry_chr22.all_assessed_pairs == "6"
    assert entry_chr22.all_switches == "0.0"
    assert entry_chr22.all_switchflips == "0.0/0.25"
    assert entry_chr22.blockwise_hamming == "0.25"
    assert entry_chr22.blockwise_diff_genotypes == "1"
    assert entry_chr22.largestblock_assessed_pairs == "4"
    assert entry_chr22.largestblock_switches == "0.0"
    assert entry_chr22.largestblock_switchflips == "0.0/0.25"
    assert entry_chr22.largestblock_hamming == "0.25"
    assert entry_chr22.largestblock_diff_genotypes == "1"


def test_compare_only_snvs(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_compare(
        vcf=["tests/data/phased1.vcf", "tests/data/phased2.vcf"],
        ploidy=2,
        names="p1,p2",
        tsv_pairwise=outtsv,
        sample="sample2",
        only_snvs=True,
    )
    with open(outtsv) as t:
        lines = [line.split("\t") for line in t]
    assert len(lines) == 3
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chrA, entry_chrB = [Fields(*l) for l in lines[1:]]

    assert entry_chrA.chromosome == "chrA"
    assert entry_chrA.all_assessed_pairs == "3"
    assert entry_chrA.all_switches == "2"
    assert entry_chrA.all_switchflips == "0/1"
    assert entry_chrA.largestblock_assessed_pairs == "3"
    assert entry_chrA.largestblock_switches == "2"
    assert entry_chrA.largestblock_hamming == "1"

    assert entry_chrB.chromosome == "chrB"
    assert entry_chrB.all_assessed_pairs == "1"
    assert entry_chrB.all_switches == "1"
    assert entry_chrB.all_switchflips == "1/0"
    assert entry_chrB.largestblock_assessed_pairs == "1"
    assert entry_chrB.largestblock_switches == "1"
    assert entry_chrB.largestblock_hamming == "1"


def test_compare_unphased():
    run_compare(
        vcf=["tests/data/unphased.vcf", "tests/data/unphased.vcf", "tests/data/unphased.vcf"],
        ploidy=2,
        sample="sample1",
    )


def test_compute_switch_flips_poly():
    phasing0 = ["0100", "1011"]
    phasing1 = ["0000", "1111"]
    sfp = compute_switch_flips_poly(phasing0, phasing1, flip_cost=3)
    assert sfp.switches == 2.0
    assert sfp.flips == 0

    phasing = ["00000000", "11111111"]
    truth = ["00000000", "11111111"]
    sfp = compute_switch_flips_poly(phasing, truth)
    assert sfp.flips + sfp.switches == 0.0

    phasing = [[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 1, 1, 1, 1], [0, 0, 0, 0, 0, 0, 0, 0]]
    sfp = compute_switch_flips_poly(phasing, truth)
    assert sfp.flips + sfp.switches == 0.0

    phasing = [[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0]]
    sfp = compute_switch_flips_poly(phasing, truth)
    assert sfp.flips + sfp.switches == 2.0

    phasing = [[1, 1, 1, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    sfp = compute_switch_flips_poly(phasing, truth)
    assert sfp.flips + sfp.switches == 1.0

    phasing = [[1, 1, 1, 1, 0, 0, 1, 0], [0, 0, 0, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    sfp = compute_switch_flips_poly(phasing, truth)
    assert sfp.flips + sfp.switches == 1.5

    phasing = [[1, 1, 1, 1, 0, 0, 1, 0], [0, 0, 0, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    sfp = compute_switch_flips_poly(phasing, truth, flip_cost=5, switch_cost=1)
    assert sfp.flips * 5 + sfp.switches == 3.5

    phasing = [[1, 1, 1, 1, 0, 0, 1, 0], [0, 0, 0, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    sfp = compute_switch_flips_poly(phasing, truth, flip_cost=1, switch_cost=10)
    assert sfp.flips + sfp.switches * 10 == 3.5

    phasing = [[0, 0, 0, 1, 0, 0, 0, 0], [1, 1, 1, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    sfp = compute_switch_flips_poly(phasing, truth)
    assert sfp.flips + sfp.switches == 1.0

    phasing = [[0, 0, 0, 1, 0, 0, 0, 0], [1, 1, 1, 0, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    sfp = compute_switch_flips_poly(phasing, truth, flip_cost=5, switch_cost=1)
    assert sfp.flips * 5 + sfp.switches == 2.0

    phasing = [[0, 0, 0, 1, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    truth = [[0, 0, 0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1, 1, 1]]
    sfp = compute_switch_flips_poly(phasing, truth, flip_cost=float("inf"), switch_cost=1)
    assert sfp.flips * float("inf") + sfp.switches == float("inf")


def test_compare_block():
    phasing = ["1111111111", "0000000000"]
    truth = ["1111100000", "0000011111"]
    phasing_errors = compare_block(phasing, truth)
    assert phasing_errors.switches == 1
    assert phasing_errors.hamming == 5

    phasing = ["000000", "101111", "111010"]
    truth = ["000000", "101010", "111111"]
    phasing_errors = compare_block(phasing, truth)
    assert phasing_errors.hamming == 2.0 / 3.0
    switch_flips = phasing_errors.switch_flips
    assert switch_flips.switches == 2.0 / 3.0

    phasing = ["1110001", "1011101", "0000010"]
    truth = ["1110001", "1010010", "0001101"]
    phasing_errors = compare_block(phasing, truth)
    assert phasing_errors.hamming == 4.0 / 3.0
    switch_flips = phasing_errors.switch_flips
    assert switch_flips.switches == 2.0 / 3.0

    phasing = ["1111101", "1010001", "0000010"]
    truth = ["1110001", "1010010", "0001101"]
    phasing_errors = compare_block(phasing, truth)
    assert phasing_errors.hamming == 6.0 / 3.0
    switch_flips = phasing_errors.switch_flips
    assert switch_flips.switches == 3.0 / 3.0

    phasing = ["111111", "111111", "111111"]
    truth = ["111111", "000000", "111111"]
    phasing_errors = compare_block(phasing, truth)
    assert phasing_errors.hamming == 2.0
    switch_flips = phasing_errors.switch_flips
    assert switch_flips.switches == 0.0


def test_compare_ignore_sample_name(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_compare(
        vcf=["tests/data/phased_single_sample1.vcf", "tests/data/phased_single_sample2.vcf"],
        ploidy=2,
        names="p1,p2",
        tsv_pairwise=outtsv,
        sample=None,
        ignore_sample_name=True,
    )
    lines = [l.split("\t") for l in open(outtsv)]
    assert len(lines) == 2
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry = Fields(*lines[1])

    assert entry.chromosome == "chrA"
    assert entry.sample == "sample1_nr1"
    assert entry.all_assessed_pairs == "3"
    assert entry.all_switches == "2"
    assert entry.all_switchflips == "2/0"
    assert entry.blockwise_hamming == "2"
    assert entry.largestblock_assessed_pairs == "2"
    assert entry.largestblock_switches == "1"
    assert entry.largestblock_hamming == "1"



# cython: language_level=3

"""
Declarations for all C++ classes that are wrapped from Cython.
"""
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libc.stdint cimport uint32_t, uint64_t
from libcpp.unordered_map cimport unordered_map


cdef extern from "../src/read.h":
    cdef cppclass Read:
        Read(string, int, int, int, int, string) except +
        Read(Read) except +
        string toString() except +
        void addVariant(int, int, int) except +
        string getName() except +
        vector[int] getMapqs() except +
        void addMapq(int) except +
        int getPosition(int) except +
        void setPosition(int, int)  except +
        int getAllele(int) except +
        void setAllele(int, int) except +
        int getVariantQuality(int) except +
        void setVariantQuality(int, int) except +
        int getVariantCount() except +
        void sortVariants() except +
        bool isSorted() except +
        int getSourceID() except +
        int getSampleID() except +
        int getReferenceStart() except +
        string getBXTag() except +
        bool hasBXTag() except +


cdef extern from "../src/indexset.h":
    cdef cppclass IndexSet:
        IndexSet() except +
        bool contains(int) except +
        void add(int) except +
        int size() except +
        string toString() except +


cdef extern from "../src/readset.h":
    cdef cppclass ReadSet:
        ReadSet() except +
        void add(Read*) except +
        string toString() except +
        int size() except +
        void sort() except +
        Read* get(int) except +
        Read* getByName(string, int) except +
        ReadSet* subset(IndexSet*) except +
        # TODO: Check why adding "except +" here doesn't compile
        vector[unsigned int]* get_positions()


cdef extern from "../src/pedigree.h":
    cdef cppclass Pedigree:
        Pedigree() except +
        void addIndividual(unsigned int id, vector[Genotype*] genotypes, vector[PhredGenotypeLikelihoods*]) except +
        void addRelationship(unsigned int f, unsigned int m, unsigned int c) except +
        unsigned int size()
        string toString() except +
        const Genotype* get_genotype_by_id(unsigned int, unsigned int) except +
        const PhredGenotypeLikelihoods* get_genotype_likelihoods_by_id(unsigned int, unsigned int) except +
        unsigned int get_variant_count() except +
        unsigned int triple_count() except +


cdef extern from "../src/pedigreedptable.h":
    cdef cppclass PedigreeDPTable:
        PedigreeDPTable(ReadSet*, vector[unsigned int], Pedigree* pedigree, bool distrust_genotypes, vector[unsigned int]* positions) except +
        void get_super_reads(vector[ReadSet*]*, vector[unsigned int]* transmission_vector) except +
        int get_optimal_score() except +
        vector[bool]* get_optimal_partitioning()
        
        
cdef extern from "../src/binomial.h":
    cdef int binomial_coefficient(int n, int k) except +
        
        
cdef extern from "../src/genotype.h":
    cdef cppclass Genotype:
        Genotype() except +
        Genotype(vector[uint32_t]) except +
        Genotype(Genotype) except +
        vector[uint32_t] as_vector() except +
        bool is_none() except +
        uint64_t get_index() except +
        string toString() except +
        bool is_homozygous() except +
        bool is_diploid_and_biallelic() except +
        uint32_t get_ploidy() except +
    cdef bool operator==(Genotype,Genotype) except +
    cdef bool operator!=(Genotype,Genotype) except +
    cdef bool operator<(Genotype,Genotype) except +
    cdef vector[uint32_t] convert_index_to_alleles(uint64_t index, uint32_t ploidy) except +
    cdef uint32_t get_max_genotype_ploidy() except +
    cdef uint32_t get_max_genotype_alleles() except +


cdef extern from "../src/genotypedptable.h":
    cdef cppclass GenotypeDPTable:
        GenotypeDPTable(ReadSet*, vector[unsigned int], Pedigree* pedigree, vector[unsigned int]* positions) except +
        vector[long double] get_genotype_likelihoods(unsigned int individual, unsigned int position) except +

cdef extern from "../src/phredgenotypelikelihoods.h":
    cdef cppclass PhredGenotypeLikelihoods:
        PhredGenotypeLikelihoods(vector[double], unsigned int, unsigned int) except +
        PhredGenotypeLikelihoods(PhredGenotypeLikelihoods) except +
        double get(Genotype) except +
        string toString() except +
        unsigned int get_ploidy() except +
        unsigned int get_nr_alleles() except +
        unsigned int size() except +
        vector[double] as_vector() except +
        void get_genotypes(vector[Genotype]&) except +


cdef extern from "../src/genotypedistribution.h":
    cdef cppclass GenotypeDistribution:
        GenotypeDistribution(double hom_ref_prob, double het_prob, double hom_alt_prob) except +
        double probabilityOf(unsigned int genotype) except +


cdef extern from "../src/genotyper.h":
    void compute_genotypes(ReadSet, vector[Genotype]* genotypes, vector[GenotypeDistribution]* genotype_likelihoods, vector[unsigned int]* positions)  except +
    void compute_polyploid_genotypes(ReadSet, size_t ploidy, vector[Genotype]* genotypes, vector[unsigned int]* positions)  except +


cdef extern from "../src/hapchat/hapchatcore.cpp":
    cdef cppclass HapChatCore:
        HapChatCore(ReadSet*)
        void get_super_reads(vector[ReadSet*]*)
        vector[bool]* get_optimal_partitioning()
        int get_length()
        int get_optimal_cost()


cdef extern from "../src/polyphase/clustereditingsolver.h":
    cdef cppclass ClusterEditingSolver:
        ClusterEditingSolver(TriangleSparseMatrix m, bool bundleEdges) except +
        ClusterEditingSolution run() except +


cdef extern from "../src/polyphase/clustereditingsolution.h":
    cdef cppclass ClusterEditingSolution:
        ClusterEditingSolution() except +
        ClusterEditingSolution(ClusterEditingSolution) except +
        ClusterEditingSolution(double pTotalCost, vector[vector[int]] pClusters) except +
        vector[unsigned int] getCluster(int i) except +
        double getTotalCost() except +
        int getNumClusters() except +


cdef extern from "../src/polyphase/trianglesparsematrix.h":
    cdef cppclass TriangleSparseMatrix:
        TriangleSparseMatrix() except +
        unsigned int entryToIndex(unsigned int i, unsigned int j) except +
        unsigned int size() except +
        float get(unsigned int i, unsigned int j) except +
        void set(unsigned int i, unsigned int j, float v) except +
        vector[pair[uint32_t, uint32_t]] getEntries() except +


cdef extern from "../src/polyphase/readscoring.h":
    cdef cppclass ReadScoring:
        ReadScoring() except +
        void scoreReadset(TriangleSparseMatrix* result, ReadSet* readset, uint32_t minOverlap, uint32_t ploidy, double err) except +


cdef extern from "../src/polyphase/haplothreader.h":
    cdef cppclass HaploThreader:
        HaploThreader(uint32_t ploidy, double switchCost, double affineSwitchCost, uint32_t maxClusterGap, uint32_t rowLimit) except +
        vector[vector[uint32_t]] computePaths(uint32_t start, uint32_t end,
                    vector[vector[uint32_t]]& covMap,
                    vector[unordered_map[uint32_t, unordered_map[uint32_t, uint32_t]]]& alleleDepths) except +
        vector[vector[uint32_t]] computePaths(vector[uint32_t]& blockStarts,
                    vector[vector[uint32_t]]& covMap,
                    vector[unordered_map[uint32_t, unordered_map[uint32_t, uint32_t]]]& alleleDepths) except +


cdef extern from "../src/polyphase/switchflipcalculator.h":
    cdef cppclass SwitchFlipCalculator:
        SwitchFlipCalculator(uint32_t ploidy, double switchCost, double flipCost) except +
        pair[double, double] compare(vector[vector[uint32_t]]& phasing0,
                    vector[vector[uint32_t]]& phasing1,
                    vector[uint32_t]& switchesInColumn,
                    vector[vector[uint32_t]]& flippedHapsInColumn,
                    vector[vector[uint32_t]]& permInColumn) except +



"""
Phase variants in a polyploid VCF using a clustering+threading algorithm.

Read a VCF and one or more files with phase information (BAM/CRAM or VCF phased
blocks) and phase the variants. The phased VCF is written to standard output.
Requires to specify a ploidy for the phasable input. Allows to specify a block
cut sensitivity to balance out length and accuracy of phased blocks.

"""
import sys
import logging
import platform
import argparse

from collections import namedtuple
from copy import deepcopy
from multiprocessing import Pool

from contextlib import ExitStack

from whatshap import __version__
from whatshap.core import (
    Read,
    ReadSet,
    Genotype,
    NumericSampleIds,
)
from whatshap.polyphase_solver import (
    ClusterEditingSolver,
    compute_polyploid_genotypes,
    scoreReadset,
)
from whatshap.cli import log_memory_usage, PhasedInputReader, CommandLineError
from whatshap.polyphaseplots import draw_plots
from whatshap.polyphaseutil import create_genotype_list, compute_block_starts, split_readset
from whatshap.reorder import run_reordering
from whatshap.threading import run_threading
from whatshap.timer import StageTimer
from whatshap.vcf import VcfReader, PhasedVcfWriter, PloidyError

__author__ = "Jana Ebler, Sven Schrinner"

PhasingParameter = namedtuple(
    "PhasingParameter",
    [
        "ploidy",
        "verify_genotypes",
        "ce_bundle_edges",
        "distrust_genotypes",
        "min_overlap",
        "block_cut_sensitivity",
        "plot_clusters",
        "plot_threading",
        "threads",
    ],
)

SingleBlockResult = namedtuple(
    "SingleBlockResult",
    ["block_id", "clustering", "paths", "cuts", "hap_cuts", "haplotypes"],
)

logger = logging.getLogger(__name__)


def run_polyphase(
    phase_input_files,
    variant_file,
    ploidy,
    reference=None,
    output=sys.stdout,
    samples=None,
    chromosomes=None,
    verify_genotypes=False,
    ignore_read_groups=False,
    indels=True,
    mapping_quality=20,
    tag="PS",
    include_haploid_sets=False,
    distrust_genotypes=False,
    write_command_line_header=True,
    read_list_filename=None,
    ce_bundle_edges=False,
    min_overlap=2,
    plot_clusters=False,
    plot_threading=False,
    block_cut_sensitivity=4,
    threads=1,
):
    """
    Run Polyploid Phasing.

    phase_input_files -- list of paths to BAM/CRAM/VCF files
    variant-file -- path to input VCF
    reference -- path to reference FASTA
    output -- path to output VCF or a file like object
    samples -- names of samples to phase. An empty list means: phase all samples
    chromosomes -- names of chromosomes to phase. An empty list means: phase all chromosomes
    ignore_read_groups
    mapping_quality -- discard reads below this mapping quality
    tag -- How to store phasing info in the VCF, can be 'PS' or 'HP'
    write_command_line_header -- whether to add a ##commandline header to the output VCF
    """
    timers = StageTimer()
    logger.info(
        "This is WhatsHap (polyploid) %s running under Python %s",
        __version__,
        platform.python_version(),
    )
    numeric_sample_ids = NumericSampleIds()
    with ExitStack() as stack:
        assert phase_input_files
        phased_input_reader = stack.enter_context(
            PhasedInputReader(
                phase_input_files,
                reference,
                numeric_sample_ids,
                ignore_read_groups,
                indels=indels,
                mapq_threshold=mapping_quality,
            )
        )
        assert not phased_input_reader.has_vcfs

        if write_command_line_header:
            command_line = "(whatshap {}) {}".format(__version__, " ".join(sys.argv[1:]))
        else:
            command_line = None
        try:
            vcf_writer = stack.enter_context(
                PhasedVcfWriter(
                    command_line=command_line,
                    in_path=variant_file,
                    out_file=output,
                    tag=tag,
                    ploidy=ploidy,
                    include_haploid_sets=include_haploid_sets,
                )
            )
        except OSError as e:
            raise CommandLineError(e)

        vcf_reader = stack.enter_context(
            VcfReader(
                variant_file, indels=indels, phases=True, genotype_likelihoods=False, ploidy=ploidy
            )
        )

        if ignore_read_groups and not samples and len(vcf_reader.samples) > 1:
            raise CommandLineError(
                "When using --ignore-read-groups on a VCF with "
                "multiple samples, --sample must also be used."
            )
        if not samples:
            samples = vcf_reader.samples

        vcf_sample_set = set(vcf_reader.samples)
        for sample in samples:
            if sample not in vcf_sample_set:
                raise CommandLineError(
                    f"Sample {sample!r} requested on command-line not found in VCF"
                )

        samples = frozenset(samples)

        read_list_file = None
        if read_list_filename:
            raise NotImplementedError("create_read_list_file not implemented")
            # read_list_file = create_read_list_file(read_list_filename)

        # Store phasing parameters in tuple to keep function signatures cleaner
        phasing_param = PhasingParameter(
            ploidy=ploidy,
            verify_genotypes=verify_genotypes,
            ce_bundle_edges=ce_bundle_edges,
            distrust_genotypes=distrust_genotypes,
            min_overlap=min_overlap,
            block_cut_sensitivity=block_cut_sensitivity,
            plot_clusters=plot_clusters,
            plot_threading=plot_threading,
            threads=threads,
        )

        timers.start("parse_vcf")
        try:
            for variant_table in vcf_reader:
                chromosome = variant_table.chromosome
                timers.stop("parse_vcf")
                if (not chromosomes) or (chromosome in chromosomes):
                    logger.info("======== Working on chromosome %r", chromosome)
                else:
                    logger.info(
                        "Leaving chromosome %r unchanged (present in VCF but not requested by option --chromosome)",
                        chromosome,
                    )
                    with timers("write_vcf"):
                        superreads, components = dict(), dict()
                        vcf_writer.write(chromosome, superreads, components)
                    continue

                # These two variables hold the phasing results for all samples
                superreads, components, haploid_components = dict(), dict(), dict()

                # Iterate over all samples to process
                for sample in samples:
                    logger.info("---- Processing individual %s", sample)

                    # Process inputs for this sample
                    missing_genotypes = set()
                    heterozygous = set()

                    genotypes = variant_table.genotypes_of(sample)
                    for index, gt in enumerate(genotypes):
                        if gt.is_none():
                            missing_genotypes.add(index)
                        elif not gt.is_homozygous():
                            heterozygous.add(index)
                        else:
                            assert gt.is_homozygous()
                    to_discard = set(range(len(variant_table))).difference(heterozygous)
                    phasable_variant_table = deepcopy(variant_table)
                    # Remove calls to be discarded from variant table
                    phasable_variant_table.remove_rows_by_index(to_discard)

                    logger.info(
                        "Number of variants skipped due to missing genotypes: %d",
                        len(missing_genotypes),
                    )
                    logger.info(
                        "Number of remaining heterozygous variants: %d", len(phasable_variant_table)
                    )

                    # Get the reads belonging to this sample
                    timers.start("read_bam")
                    readset, vcf_source_ids = phased_input_reader.read(
                        chromosome, phasable_variant_table.variants, sample
                    )
                    readset.sort()
                    timers.stop("read_bam")

                    # Verify genotypes
                    if verify_genotypes:
                        timers.start("verify_genotypes")
                        logger.info("Verify genotyping of %s", sample)
                        positions = [v.position for v in phasable_variant_table.variants]
                        computed_genotypes = [
                            Genotype(gt)
                            for gt in compute_polyploid_genotypes(readset, ploidy, positions)
                        ]
                        # skip all positions at which genotypes do not match
                        given_genotypes = phasable_variant_table.genotypes_of(sample)
                        matching_genotypes = []
                        missing_genotypes = set()
                        for i, g in enumerate(given_genotypes):
                            c_g = computed_genotypes[i]
                            if (g == c_g) or (c_g is None):
                                matching_genotypes.append(g)
                            else:
                                matching_genotypes.append(Genotype([]))
                                missing_genotypes.add(i)
                        phasable_variant_table.set_genotypes_of(sample, matching_genotypes)

                        # Remove variants with deleted genotype
                        phasable_variant_table.remove_rows_by_index(missing_genotypes)
                        logger.info(
                            "Number of variants removed due to inconsistent genotypes: %d",
                            len(missing_genotypes),
                        )
                        logger.info(
                            "Number of remaining heterozygous variants: %d",
                            len(phasable_variant_table),
                        )

                        # Re-read the readset to remove discarded variants
                        readset, vcf_source_ids = phased_input_reader.read(
                            chromosome, phasable_variant_table.variants, sample
                        )
                        readset.sort()
                        timers.stop("verify_genotypes")

                    # Remove reads with insufficient variants
                    readset = readset.subset(
                        [i for i, read in enumerate(readset) if len(read) >= max(2, min_overlap)]
                    )
                    logger.info("Kept %d reads that cover at least two variants each", len(readset))

                    # Adapt the variant table to the subset of reads
                    phasable_variant_table.subset_rows_by_position(readset.get_positions())

                    # Run the actual phasing
                    (
                        sample_components,
                        sample_haploid_components,
                        sample_superreads,
                    ) = phase_single_individual(
                        readset, phasable_variant_table, sample, phasing_param, output, timers
                    )

                    # Collect results
                    components[sample] = sample_components
                    haploid_components[sample] = sample_haploid_components
                    superreads[sample] = sample_superreads

                with timers("write_vcf"):
                    logger.info("======== Writing VCF")
                    vcf_writer.write(
                        chromosome,
                        superreads,
                        components,
                        haploid_components if include_haploid_sets else None,
                    )
                    logger.info("Done writing VCF")
                logger.debug("Chromosome %r finished", chromosome)
                timers.start("parse_vcf")
            timers.stop("parse_vcf")
        except PloidyError as e:
            raise CommandLineError(e)

    if read_list_file:
        read_list_file.close()

    logger.info("\n== SUMMARY ==")

    log_memory_usage(include_children=(threads > 1))
    logger.info("Time spent reading BAM/CRAM:         %6.1f s", timers.elapsed("read_bam"))
    logger.info("Time spent parsing VCF:              %6.1f s", timers.elapsed("parse_vcf"))
    if verify_genotypes:
        logger.info(
            "Time spent verifying genotypes:      %6.1f s", timers.elapsed("verify_genotypes")
        )
    logger.info("Time spent detecting blocks:         %6.1f s", timers.elapsed("detecting_blocks"))
    if threads == 1:
        logger.info("Time spent scoring reads:            %6.1f s", timers.elapsed("read_scoring"))
        logger.info("Time spent solving cluster editing:  %6.1f s", timers.elapsed("clustering"))
        logger.info("Time spent threading haplotypes:     %6.1f s", timers.elapsed("threading"))
        logger.info("Time spent reordering haplotypes:    %6.1f s", timers.elapsed("reordering"))
    else:
        """
        TODO: The runtime measurement for the different stages does not properly for multithreading,
        because the global timer is not visible from within the phase_single_block_mt method.
        Workaround is to only report the total phasing time.
        """
        logger.info("Time spent phasing blocks:           %6.1f s", timers.elapsed("phase_blocks"))
    if plot_clusters or plot_threading:
        logger.info("Time spent creating plots:           %6.1f s", timers.elapsed("create_plots"))
    logger.info("Time spent writing VCF:              %6.1f s", timers.elapsed("write_vcf"))
    logger.info("Time spent on rest:                  %6.1f s", timers.total() - timers.sum())
    logger.info("Total elapsed time:                  %6.1f s", timers.total())


def phase_single_individual(readset, phasable_variant_table, sample, param, output, timers):

    # Compute the genotypes that belong to the variant table and create a list of all genotypes
    genotype_list = create_genotype_list(phasable_variant_table, sample)

    # Precompute block borders based on read coverage and linkage between variants
    logger.info("Detecting connected components with weak interconnect ..")
    timers.start("detecting_blocks")
    num_vars = len(readset.get_positions())
    ploidy = param.ploidy
    if param.block_cut_sensitivity == 0:
        block_starts = [0]
    elif param.block_cut_sensitivity == 1:
        block_starts = compute_block_starts(readset, ploidy, single_linkage=True)
    else:
        block_starts = compute_block_starts(readset, ploidy, single_linkage=False)

    # Set block borders and split readset
    block_starts.append(num_vars)
    num_non_singleton_blocks = len(
        [i for i in range(len(block_starts) - 1) if block_starts[i] < block_starts[i + 1] - 1]
    )
    logger.info(
        "Split heterozygous variants into {} blocks (and {} singleton blocks).".format(
            num_non_singleton_blocks, len(block_starts) - num_non_singleton_blocks - 1
        )
    )

    block_readsets = split_readset(readset, block_starts)
    timers.stop("detecting_blocks")

    # Process blocks independently
    results = []

    # Create genotype slices for blocks
    gt_slices = []
    for block_id, block_readset in enumerate(block_readsets):
        block_start = block_starts[block_id]
        block_end = block_starts[block_id + 1]
        block_num_vars = block_end - block_start

        assert len(block_readset.get_positions()) == block_num_vars
        gt_slices.append(genotype_list[block_start:block_end])

    processed_non_singleton_blocks = 0

    """
    Python's multiprocessing makes hard copies of the passed arguments, which is not trivial for
    cython objects, especially when they contain pointers to other cython objects. Any passed
    object must be (de)serializable (in Python: pickle). All other objects created in the main
    thread are also accessible by the workers, but they are handled via the copy-on-write policy.
    This means, that e.g. the large main readset is not hardcopied for every thread, as long as it
    is not modified there. Since this would cause a massive waste of memory, this must not be done
    and the main readset must also never be passed as argument to the workers.
    """
    if param.threads == 1:
        # for single-threading, process everything individually to minimize memory footprint
        for block_id, block_readset in enumerate(block_readsets):
            block_num_vars = block_starts[block_id + 1] - block_starts[block_id]
            if block_num_vars > 1:
                processed_non_singleton_blocks += 1
                logger.info(
                    "Processing block {} of {} with {} reads and {} variants.".format(
                        processed_non_singleton_blocks,
                        num_non_singleton_blocks,
                        len(block_readset),
                        block_num_vars,
                    )
                )
            results.append(
                phase_single_block(block_id, block_readset, gt_slices[block_id], param, timers)
            )

    else:
        # sort block by descending size (4/3-approximation for scheduling problem)
        joblist = [(i, len(block_readsets[i])) for i in range(len(block_readsets))]
        joblist.sort(key=lambda x: -x[1])

        timers.start("phase_blocks")
        with Pool(processes=param.threads) as pool:
            process_results = [
                pool.apply_async(
                    phase_single_block_mt,
                    (
                        block_id,
                        block_readsets[block_id],
                        gt_slices[block_id],
                        param,
                        timers,
                        job_id,
                        num_non_singleton_blocks,
                    ),
                )
                for job_id, (block_id, block_readset) in enumerate(joblist)
            ]
            # collect all blockwise results
            blockwise_results = [res.get() for res in process_results]
            results = sorted(blockwise_results, key=lambda x: x.block_id)

        timers.stop("phase_blocks")

    # Aggregate blockwise results
    clustering, threading, haplotypes, cuts, hap_cuts = aggregate_results(results, ploidy)

    # Summarize data for VCF file
    accessible_pos = sorted(readset.get_positions())
    components = {}
    haploid_components = {}

    cuts = cuts + [num_vars]
    for i, cut_pos in enumerate(cuts[:-1]):
        for pos in range(cuts[i], cuts[i + 1]):
            components[accessible_pos[pos]] = accessible_pos[cuts[i]]
            components[accessible_pos[pos] + 1] = accessible_pos[cuts[i]]
            haploid_components[accessible_pos[pos]] = [0] * ploidy
            haploid_components[accessible_pos[pos] + 1] = [0] * ploidy

    for j in range(ploidy):
        hap_cuts[j] = hap_cuts[j] + [num_vars]
        for i, cut_pos in enumerate(hap_cuts[j][:-1]):
            for pos in range(hap_cuts[j][i], hap_cuts[j][i + 1]):
                haploid_components[accessible_pos[pos]][j] = accessible_pos[hap_cuts[j][i]]
                haploid_components[accessible_pos[pos] + 1][j] = accessible_pos[hap_cuts[j][i]]

    superreads = ReadSet()
    for i in range(ploidy):
        read = Read(f"superread {i + 1}", 0, 0)
        # insert alleles
        for j, allele in enumerate(haplotypes[i]):
            if allele < 0:
                continue
            read.add_variant(accessible_pos[j], allele, 0)
        superreads.add(read)

    # Plot option
    if param.plot_clusters or param.plot_threading:
        timers.start("create_plots")
        draw_plots(
            block_readsets,
            clustering,
            threading,
            haplotypes,
            cuts[:-1],
            phasable_variant_table,
            param.plot_clusters,
            param.plot_threading,
            output,
        )
        timers.stop("create_plots")

    # Return results
    return components, haploid_components, superreads


def phase_single_block(block_id, block_readset, genotype_slice, param, timers):
    """
    Takes as input data the reads from a single (pre-computed) block and the genotypes for all
    variants inside the block. Runs a three-phase algorithm to compute a phasing for this isolated
    block. Input are four objects:

    block_readset -- Input reads
    genotype_slice -- Genotype (as dictionary) for every position
    param -- Object containing phasing parameters
    timers -- Timer object to measure time
    """

    block_num_vars = len(block_readset.get_positions())

    # Check for singleton blocks and handle them differently (for efficiency reasons)
    if block_num_vars == 1:

        # construct trivial solution for singleton blocks, by using the genotype as phasing
        allele_to_id = dict()
        for allele in genotype_slice[0]:
            if allele not in allele_to_id:
                allele_to_id[allele] = len(allele_to_id)
        clusts = [[] for _ in range(len(allele_to_id))]
        for i, read in enumerate(block_readset):
            clusts[allele_to_id[read[0].allele]].append(i)

        paths = [[]]
        haps = []
        for allele in genotype_slice[0]:
            for i in range(genotype_slice[0][allele]):
                paths[0].append(allele_to_id[allele])
                haps.append([allele])

        return SingleBlockResult(
            block_id, clusts, paths, [0], [[0] for _ in range(param.ploidy)], haps
        )

    # Block is non-singleton here, so run the normal routine
    # Phase I: Cluster Editing

    # Compute similarity values for all read pairs
    timers.start("read_scoring")
    logger.debug("Computing similarities for read pairs ..")
    sim = scoreReadset(block_readset, param.min_overlap, param.ploidy, 0.07)

    # Run cluster editing
    logger.debug(
        "Solving cluster editing instance with {} nodes and {} edges ..".format(
            len(block_readset), len(sim)
        )
    )
    timers.stop("read_scoring")
    timers.start("clustering")
    solver = ClusterEditingSolver(sim, param.ce_bundle_edges)
    clustering = solver.run()
    del solver
    del sim

    # Add trailing isolated nodes to single-ton clusters, if missing
    nodes_in_c = sum(len(c) for c in clustering)
    for i in range(nodes_in_c, len(block_readset)):
        clustering.append([i])

    timers.stop("clustering")

    # Phase II: Threading

    # Assemble clusters to haplotypes
    logger.debug(f"Threading haplotypes through {len(clustering)} clusters ..\r")
    timers.start("threading")

    # Add dynamic programming for finding the most likely subset of clusters
    paths, haplotypes = run_threading(
        block_readset,
        clustering,
        param.ploidy,
        genotypes=None if param.distrust_genotypes else genotype_slice,
    )
    timers.stop("threading")

    # Phase III: Reordering

    logger.debug("Reordering ambiguous sites ..\r")
    timers.start("reordering")
    cut_positions, haploid_cuts, path, haplotypes = run_reordering(
        block_readset, clustering, paths, haplotypes, param.block_cut_sensitivity
    )
    timers.stop("reordering")

    # collect results from threading
    return SingleBlockResult(
        block_id=block_id,
        clustering=clustering,
        paths=paths,
        cuts=cut_positions,
        hap_cuts=haploid_cuts,
        haplotypes=haplotypes,
    )


def phase_single_block_mt(
    block_id, block_readset, genotype_slice, param, timers, job_id, num_blocks
):
    """
    Wrapper for the phase_single_block() function. Carries a block_id through to the results
    and unpythonizes the given readset, because cython objects are very troublesome to hardcopy.
    """
    block_vars = len(block_readset.get_positions())
    if block_vars > 1:
        logger.info(
            "Phasing block {} of {} with {} reads and {} variants.".format(
                job_id + 1, num_blocks, len(block_readset), block_vars
            )
        )
    result = phase_single_block(block_id, block_readset, genotype_slice, param, timers)
    del block_readset
    if block_vars > 1:
        logger.info(f"Finished block {job_id + 1}.")
    return result


def aggregate_results(results, ploidy):
    """
    Collects all blockwise phasing results and aggregates them into one list for each type of
    information. Local ids and indices are converted to globals ones in this step.
    """

    clustering, cuts = [], []
    paths = []
    hap_cuts = [[] for _ in range(ploidy)]
    haplotypes = [[] for _ in range(ploidy)]
    rid_offset, cid_offset, pos_offset = 0, 0, 0
    for r in results:
        clustering += [[rid_offset + rid for rid in clust] for clust in r.clustering]
        paths += [[cid_offset + cid for cid in p] for p in r.paths]
        cuts += [pos_offset + c for c in r.cuts]
        for hap_cut, ext in zip(hap_cuts, r.hap_cuts):
            hap_cut += [pos_offset + h for h in ext]
        for hap, ext in zip(haplotypes, r.haplotypes):
            hap += ext
        rid_offset = max(rid for clust in clustering for rid in clust)
        cid_offset = len(clustering)
        pos_offset = len(haplotypes[0])

    return clustering, paths, haplotypes, cuts, hap_cuts


def add_arguments(parser):
    arg = parser.add_argument
    # Positional argument
    arg(
        "variant_file",
        metavar="VCF",
        help="VCF file with variants to be phased (can be gzip-compressed)",
    )
    arg(
        "phase_input_files",
        nargs="*",
        metavar="PHASEINPUT",
        help="BAM or CRAM with sequencing reads.",
    )
    arg(
        "-o",
        "--output",
        default=sys.stdout,
        help="Output VCF file. Add .gz to the file name to get compressed output. "
        "If omitted, use standard output.",
    )
    arg(
        "--reference",
        "-r",
        metavar="FASTA",
        help="Reference file. Provide this to detect alleles through re-alignment. "
        "If no index (.fai) exists, it will be created",
    )
    arg(
        "--tag",
        choices=("PS", "HP"),
        default="PS",
        help="Store phasing information with PS tag (standardized) or "
        "HP tag (used by GATK ReadBackedPhasing) (default: %(default)s)",
    )
    # arg(
    #    "--output-read-list",
    #    metavar="FILE",
    #    default=None,
    #    dest="read_list_filename",
    #    help="Write reads that have been used for phasing to FILE.",
    # )

    arg = parser.add_argument_group("Input pre-processing, selection, and filtering").add_argument
    arg(
        "--mapping-quality",
        "--mapq",
        metavar="QUAL",
        default=20,
        type=int,
        help="Minimum mapping quality (default: %(default)s)",
    )
    arg(
        "--indels",
        dest="indels",
        default=False,
        action="store_true",
        help="Also phase indels (default: do not phase indels)",
    )
    arg(
        "--ignore-read-groups",
        default=False,
        action="store_true",
        help="Ignore read groups in BAM/CRAM header and assume all reads come "
        "from the same sample.",
    )
    arg(
        "--include-haploid-sets",
        default=False,
        action="store_true",
        help="Include the phase set information for every single haplotype in a custom VCF format field 'HS'.",
    )
    arg(
        "--sample",
        dest="samples",
        metavar="SAMPLE",
        default=[],
        action="append",
        help="Name of a sample to phase. If not given, all samples in the "
        "input VCF are phased. Can be used multiple times.",
    )
    arg(
        "--chromosome",
        dest="chromosomes",
        metavar="CHROMOSOME",
        default=[],
        action="append",
        help="Name of chromosome to phase. If not given, all chromosomes in the "
        "input VCF are phased. Can be used multiple times.",
    )
    arg(
        "--distrust-genotypes",
        dest="distrust_genotypes",
        action="store_true",
        default=False,
        help="Allows the phaser to change genotypes if beneficial for the internal model.",
    )

    # add polyphase specific arguments
    arg = parser.add_argument_group("Parameters for phasing steps").add_argument
    arg(
        "--ploidy",
        "-p",
        metavar="PLOIDY",
        type=int,
        required=True,
        help="The ploidy of the sample(s). Argument is required.",
    )
    arg(
        "--min-overlap",
        metavar="OVERLAP",
        type=int,
        default=2,
        help="Minimum required read overlap for internal read clustering stage (default: %(default)s).",
    )
    arg(
        "--block-cut-sensitivity",
        "-B",
        metavar="SENSITIVITY",
        type=int,
        dest="block_cut_sensitivity",
        default=4,
        help="Strategy to determine block borders. 0 yields the longest blocks with more switch errors, 5 has the shortest blocks with lowest switch error rate (default: %(default)s).",
    )
    arg(
        "--threads",
        "-t",
        metavar="THREADS",
        type=int,
        default=1,
        help="Maximum number of CPU threads used (default: %(default)s).",
    )

    # more arguments, which are experimental or for debugging and should not be presented to the user
    arg(
        "--ce-bundle-edges",
        dest="ce_bundle_edges",
        default=False,
        action="store_true",
        help=argparse.SUPPRESS,
    )  # help='Influences the cluster editing heuristic. Only for debug/developing purpose (default: %(default)s).')
    arg(
        "--plot-clusters",
        dest="plot_clusters",
        default=False,
        action="store_true",
        help=argparse.SUPPRESS,
    )  # help='Plot a super heatmap for the computed clustering (default: %(default)s).')
    arg(
        "--plot-threading",
        dest="plot_threading",
        default=False,
        action="store_true",
        help=argparse.SUPPRESS,
    )  # help='Plot the haplotypes\' threading through the read clusters (default: %(default)s).')
    arg(
        "--verify-genotypes",
        default=False,
        action="store_true",
        help=argparse.SUPPRESS,
    )  # help="Verify input genotypes by re-typing them using the given reads.",


def validate(args, parser):
    if args.block_cut_sensitivity > 5 or args.block_cut_sensitivity < 0:
        parser.error("Block cut sensitivity must be an integer value between 0 and 5.")


def main(args):
    run_polyphase(**vars(args))



from collections import defaultdict
import gzip
import logging
from typing import Optional, DefaultDict
import os
import stat
import sys

import pyfaidx
from dataclasses import dataclass


class FastaNotIndexedError(Exception):
    pass


class InvalidRegion(Exception):
    pass


def detect_file_format(path):
    """
    Detect file format and return 'BAM', 'CRAM', 'VCF' or None. None indicates an
    unrecognized file format.

    'VCF' is returned for both uncompressed and compressed VCFs (.vcf and .vcf.gz).
    """
    with open(path, "rb") as f:
        first_bytes = f.read(16)
        if first_bytes.startswith(b"CRAM"):
            return "CRAM"
        if first_bytes.startswith(b"##fileformat=VCF"):
            return "VCF"

    gzip_header = b"\037\213"
    if first_bytes.startswith(gzip_header):
        with gzip.GzipFile(path, "rb") as f:
            first_bytes = f.read(16)
            if first_bytes.startswith(b"BAM\1"):
                return "BAM"
            elif first_bytes.startswith(b"##fileformat=VCF"):
                return "VCF"

    return None


def stdout_is_regular_file() -> bool:
    """
    Detect if standard output is a regular file (or say a pipe).

    :return: True if stdout is a regular file, else False.
    """
    mode = os.fstat(sys.stdout.buffer.fileno()).st_mode
    return stat.S_ISREG(mode)


def IndexedFasta(path):
    try:
        f = pyfaidx.Fasta(path, as_raw=True, sequence_always_upper=True, build_index=False)
    except pyfaidx.IndexNotFoundError:
        raise FastaNotIndexedError(path)
    return f


def plural_s(n: int) -> str:
    return "" if n == 1 else "s"


@dataclass
class Region:
    chromosome: str
    start: int
    end: Optional[int]

    def __repr__(self):
        return f'Region("{self.chromosome}", {self.start}, {self.end})'

    @staticmethod
    def parse(spec: str):
        """
        >>> Region.parse("chr1")
        Region("chr1", 0, None)
        >>> Region.parse("chr1:")
        Region("chr1", 0, None)
        >>> Region.parse("chr1:101")
        Region("chr1", 100, None)
        >>> Region.parse("chr1:101-")
        Region("chr1", 100, None)
        >>> Region.parse("chr1:101-200")
        Region("chr1", 100, 200)
        >>> Region.parse("chr1:101:200")  # for backwards compatibility
        Region("chr1", 100, 200)
        """
        parts = spec.split(":", maxsplit=1)
        chromosome = parts[0]
        if len(parts) == 1 or not parts[1]:
            start, end = 0, None
        else:
            try:
                sep = ":" if ":" in parts[1] else "-"
                start_end = parts[1].split(sep, maxsplit=1)
                start = int(start_end[0]) - 1
                if len(start_end) == 1 or not start_end[1]:
                    end = None
                else:
                    end = int(start_end[1])
                    if end <= start:
                        raise InvalidRegion("end is before start in specified region")
            except ValueError:
                raise InvalidRegion("Region must be specified as chrom[:start[-end]])") from None
        return Region(chromosome, start, end)


_warning_count: DefaultDict[str, int] = defaultdict(int)


def warn_once(logger, msg: str, *args) -> None:
    if _warning_count[msg] == 0 and not logger.isEnabledFor(logging.DEBUG):
        logger.warning(msg + " Hiding further warnings of this type, use --debug to show", *args)
    else:
        logger.debug(msg, *args)
    _warning_count[msg] += 1



"""
Tests for 'whatshap stats'
"""
from collections import namedtuple
from whatshap.cli.stats import run_stats


def test_stats1(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_stats(
        vcf="tests/data/phased1.vcf",
        tsv=outtsv,
        sample="sample1",
        chr_lengths="tests/data/chr-lengths.txt",
    )
    with open(outtsv) as f:
        lines = [l.split("\t") for l in f]
    assert len(lines) == 4
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chrA, entry_chrB, entry_all = [Fields(*l) for l in lines[1:]]

    assert entry_chrA.chromosome == "chrA"
    assert entry_chrA.variants == "8"
    assert entry_chrA.phased == "7"
    assert entry_chrA.unphased == "1"
    assert entry_chrA.blocks == "2"
    assert entry_chrA.variant_per_block_sum == "7"
    assert entry_chrA.bp_per_block_sum == "551"
    assert entry_chrA.block_n50[:-1] == "101"

    assert entry_chrB.chromosome == "chrB"
    assert entry_chrB.variants == "2"
    assert entry_chrB.phased == "2"
    assert entry_chrB.unphased == "0"
    assert entry_chrB.blocks == "1"
    assert entry_chrB.bp_per_block_sum == "50"
    assert entry_chrB.variant_per_block_sum == "2"
    assert entry_chrB.block_n50[:-1] == "0"

    assert entry_all.chromosome == "ALL"
    assert entry_all.variants == "10"
    assert entry_all.phased == "9"
    assert entry_all.unphased == "1"
    assert entry_all.blocks == "3"
    assert entry_all.bp_per_block_sum == "601"
    assert entry_all.variant_per_block_sum == "9"
    assert entry_all.block_n50[:-1] == "0"


def test_stats2(tmp_path):
    outtsv = tmp_path / "output.tsv"
    run_stats(
        vcf="tests/data/phased3.vcf",
        tsv=outtsv,
        sample="sample1",
        chr_lengths="tests/data/chr-lengths.txt",
    )
    with open(outtsv) as f:
        lines = [l.split("\t") for l in f]
    assert len(lines) == 4
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry_chrA, entry_chrB, entry_all = [Fields(*l) for l in lines[1:]]

    assert entry_chrA.chromosome == "chrA"
    assert entry_chrA.variants == "9"
    assert entry_chrA.phased == "4"
    assert entry_chrA.unphased == "5"
    assert entry_chrA.blocks == "1"
    assert entry_chrA.variant_per_block_sum == "4"
    assert entry_chrA.bp_per_block_sum == "350"
    assert entry_chrA.block_n50[:-1] == "0"

    assert entry_chrB.chromosome == "chrB"
    assert entry_chrB.variants == "4"
    assert entry_chrB.phased == "4"
    assert entry_chrB.unphased == "0"
    assert entry_chrB.blocks == "1"
    assert entry_chrB.variant_per_block_sum == "4"
    assert entry_chrB.bp_per_block_sum == "400"
    assert entry_chrB.block_n50[:-1] == "400"

    assert entry_all.chromosome == "ALL"
    assert entry_all.variants == "13"
    assert entry_all.phased == "8"
    assert entry_all.unphased == "5"
    assert entry_all.blocks == "2"
    assert entry_all.variant_per_block_sum == "8"
    assert entry_all.bp_per_block_sum == "750"
    assert entry_all.block_n50[:-1] == "350"


def test_overlapping_phaseblocks(tmp_path):
    """
    We have three phaseblocks on chrA which is 1000 bp

        chrA:100-700 --> 600 bp
        chrA:410-470 --> 60 bp
        chrA:800-950 --> 150 bp

    Total block sum should be 600 + 60 + 150 = 810 bp

    For NG50 the first block is split since the second block overlaps, now we have four blocks

        chrA:100-350 --> 250 bp
        chrA:410-470 --> 60 bp
        chrA:500-700 --> 200 bp
        chrA:800-950 --> 150 bp

    Half of the total length is 1000 * 0.5 = 500 bp.
    Let's calculate NG50 by adding block lengths in descending order until we exceed 500 bp

        block   length  total   >500
        1       250     250     no
        2       200     450     no
        3       150     600     yes ->  NG50 = 150 bp
    """

    outtsv = tmp_path / "output.tsv"
    run_stats(
        vcf="tests/data/phased_overlapping.vcf",
        tsv=outtsv,
        sample="sample1",
    )
    with open(outtsv) as f:
        lines = [l.split("\t") for l in f]
    assert len(lines) == 2
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    entry = Fields(*lines[1])

    assert entry.chromosome == "chrA"
    assert entry.blocks == "3"
    assert entry.bp_per_block_sum == "810"
    assert entry.block_n50[:-1] == "150"



"""
Test ReadScoring
"""

from whatshap.core import Read, ReadSet
from whatshap.polyphase_solver import scoreReadset


def test_readscoring_toy():
    readset = ReadSet()
    read1 = Read("name1", 15)
    read1.add_variant(0, 0, 1)
    read1.add_variant(1, 0, 1)
    read1.add_variant(2, 0, 1)
    read1.add_variant(3, 1, 1)
    readset.add(read1)
    read2 = Read("name2", 15)
    read2.add_variant(1, 1, 1)
    read2.add_variant(2, 0, 1)
    read2.add_variant(3, 0, 1)
    read2.add_variant(4, 1, 1)
    readset.add(read2)
    read3 = Read("name3", 15)
    read3.add_variant(2, 0, 1)
    read3.add_variant(3, 1, 1)
    read3.add_variant(4, 0, 1)
    read3.add_variant(5, 1, 1)
    readset.add(read3)
    read4 = Read("name4", 15)
    read4.add_variant(3, 0, 1)
    read4.add_variant(4, 1, 1)
    read4.add_variant(5, 0, 1)
    read4.add_variant(6, 0, 1)
    readset.add(read4)
    read5 = Read("name5", 15)
    read5.add_variant(4, 0, 1)
    read5.add_variant(5, 1, 1)
    read5.add_variant(6, 1, 1)
    read5.add_variant(7, 0, 1)
    readset.add(read5)
    read6 = Read("name6", 15)
    read6.add_variant(5, 0, 1)
    read6.add_variant(6, 0, 1)
    read6.add_variant(7, 0, 1)
    read6.add_variant(8, 1, 1)
    readset.add(read6)
    read7 = Read("name7", 15)
    read7.add_variant(6, 1, 1)
    read7.add_variant(7, 0, 1)
    read7.add_variant(8, 0, 1)
    read7.add_variant(9, 1, 1)
    readset.add(read7)
    sim = scoreReadset(readset, 2, 2)

    assert sim.get(0, 1) < 0.0
    assert sim.get(0, 2) > 0.0
    assert sim.get(0, 3) <= 0.0
    assert sim.get(0, 4) >= 0.0
    assert sim.get(0, 5) <= 0.0
    assert sim.get(0, 6) >= 0.0
    assert sim.get(1, 2) < 0.0
    assert sim.get(1, 3) > 0.0
    assert sim.get(1, 4) <= 0.0
    assert sim.get(1, 5) >= 0.0
    assert sim.get(1, 6) <= 0.0
    assert sim.get(2, 3) < 0.0
    assert sim.get(2, 4) > 0.0
    assert sim.get(2, 5) <= 0.0
    assert sim.get(2, 6) >= 0.0
    assert sim.get(3, 4) < 0.0
    assert sim.get(3, 5) > 0.0
    assert sim.get(3, 6) <= 0.0
    assert sim.get(4, 5) < 0.0
    assert sim.get(4, 6) > 0.0
    assert sim.get(5, 6) < 0.0



# cython: language_level=3

from libcpp.vector cimport vector
from libcpp.unordered_map cimport unordered_map
from libcpp.pair cimport pair
from libcpp cimport bool

ctypedef int item_type
ctypedef vector[int] priority_type
ctypedef priority_type* priority_type_ptr
ctypedef pair[priority_type_ptr,item_type] queue_entry_type


cdef class PriorityQueue:
	cdef vector[queue_entry_type] heap
	cdef unordered_map[item_type,int] positions

	cdef void c_push(self, priority_type_ptr score, int item)
	cdef void _swap(self, int index1, int index2)
	cdef bool _score_lower(self, int index1, int index2)
	cdef void _sift_up(self, int index)
	cdef void _sift_down(self, int index)
	cdef queue_entry_type c_pop(self)
	cdef void c_change_score(self, item_type item, priority_type_ptr c_new_score)
	cdef priority_type_ptr c_get_score_by_item(self, item_type item)
	cdef int size(self)
	cdef bool c_is_empty(self)



from whatshap.__main__ import main

import pytest


def test_version():
    with pytest.raises(SystemExit) as exc:
        main(["--version"])
    assert exc.value.code == 0


def test_help():
    with pytest.raises(SystemExit) as exc:
        main(["--help"])
    assert exc.value.code == 0



from pytest import raises
from whatshap.bam import (
    SampleBamReader,
    SampleNotFoundError,
    AlignmentFileNotIndexedError,
)


def test_read():
    sbr = SampleBamReader("tests/data/oneread.bam")
    reads = list(sbr.fetch("ref", "sample"))
    assert len(reads) == 1
    read = reads[0]
    assert read.bam_alignment.opt("RG") == "1"


def test_read_sample_not_found():
    sbr = SampleBamReader("tests/data/oneread.bam")
    with raises(SampleNotFoundError):
        list(sbr.fetch("ref", "non-existing-sample"))


def test_read_cram():
    sbr = SampleBamReader("tests/data/oneread.cram", reference="tests/data/oneread-ref.fasta")
    reads = list(sbr.fetch("ref", "sample"))
    assert len(reads) == 1
    assert reads[0].bam_alignment.opt("RG") == "1"


def test_no_index():
    with raises(AlignmentFileNotIndexedError):
        SampleBamReader("tests/data/not-indexed.bam")
